/*******************************************************************************

  Intel Data Center Bridging (DCB) Software
  Copyright(c) 2007-2008 Intel Corporation.

  This program is free software; you can redistribute it and/or modify it
  under the terms and conditions of the GNU General Public License,
  version 2, as published by the Free Software Foundation.

  This program is distributed in the hope it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.

  The full GNU General Public License is included in this distribution in
  the file called "COPYING".

  Contact Information:
  e1000-eedc Mailing List <e1000-eedc@lists.sourceforge.net>
  Intel Corporation, 5200 N.E. Elam Young Parkway, Hillsboro, OR 97124-6497

*******************************************************************************/

#include "dcb_protocol.h"
#include "dcb_driver_interface.h"
#include "dcb_osdep.h"
#include "dcb_persist_store.h"
#include "dcb_rule_chk.h"
#include "dcb_events.h"
#include "messages.h"

#include <map>
#include <string>
boolean_t bLLDPPresent = TRUE;

/* Redefine with default specified */
dcb_result run_feature_protocol(char *device_name, u32 EventFlag,
				u32 Subtype = 0);
dcb_result run_dcb_protocol(char *device_name, u32 EventFlag, u32 Subtype = 0);

using namespace std;

typedef map <string, pg_attribs*> pg_store1;
typedef map <string, pg_attribs*>::iterator pg_it;
pg_store1 pg;
pg_store1 peer_pg;
pg_store1 oper_pg;

typedef map <string, pfc_attribs*> pfc_store;
typedef map <string, pfc_attribs*>::iterator pfc_it;
pfc_store pfc;
pfc_store peer_pfc;
pfc_store oper_pfc;

typedef map <string, bwg_descriptions*> bwg_desc_store;
typedef map <string, bwg_descriptions*>::iterator bwg_desc_it;
bwg_desc_store bwg_desc;

/* String here is adapter name appened with APP TLV subtype. */
typedef map <string, app_attribs*> app_store;
typedef map <string, app_attribs*>::iterator app_it;
app_store apptlv;
app_store peer_apptlv;
app_store oper_apptlv;


typedef map <string, control_protocol_attribs*> dcb_control_protocol;
typedef map <string, control_protocol_attribs*>::iterator control_prot_it;
dcb_control_protocol dcb_control_prot;
dcb_control_protocol dcb_peer_control_prot;


extern "C" int add_port(const char *device_name);
extern "C" int remove_port(const char *device_name);
extern "C" void process_somethingChangedLocal(char *ifname,
						boolean_t NewFeatureTLVs);


/* Add the store pointer to init_pg, i.e. memset store to 0,
 * then copy attribs to store
 */
void init_pg(pg_attribs *Attrib, pg_attribs *Store)
{
	memset(Store,0,sizeof(*Store));

	Attrib->protocol.Max_version = DCB_PG_MAX_VERSION;
	Attrib->protocol.State = DCB_INIT;
	Attrib->protocol.Advertise_prev = FALSE;
	Attrib->protocol.tlv_sent = FALSE;

	memcpy(Store, Attrib, sizeof (*Attrib));
}

/* pass in the pointer to attrib */
boolean_t add_pg(char *device_name, pg_attribs *Attrib)
{
	TRY_BEGIN;
	string dn(device_name); 

	pg_it it = pg.find(dn);
	if (it == pg.end()) {
		/* Device not present: add */
		pg_attribs *store =
			(pg_attribs*)malloc(sizeof(*store));
		if (!store) return FALSE;

		init_pg(Attrib, store);
		pg.insert(make_pair(dn, store));
	}
	else {  /* already in data store, just update it */
		(*it).second->protocol.Advertise_prev =
				(*it).second->protocol.Advertise;
		(*it).second->protocol.Advertise = Attrib->protocol.Advertise;
		(*it).second->protocol.Enable = Attrib->protocol.Enable;
		(*it).second->protocol.Willing = Attrib->protocol.Willing;

		memcpy(&(*it).second->rx, &(Attrib->rx), sizeof(Attrib->rx));
		memcpy(&(*it).second->tx, &(Attrib->tx), sizeof(Attrib->tx));
	}
	return TRUE;

	CATCH_END(FALSE);
}

void init_oper_pg(pg_attribs *Attrib)
{
	memset(Attrib,0,sizeof(*Attrib));
}

boolean_t add_oper_pg(char *device_name)
{
	TRY_BEGIN;
	string dn(device_name); 

	pg_it it = oper_pg.find(dn);
	if (it == oper_pg.end()) {
		/* Device not present: add */
		pg_attribs *store =
			(pg_attribs*)malloc(sizeof(*store));

		if (!store) return FALSE;

		init_oper_pg( store);
		oper_pg.insert(make_pair(dn, store));
	}
	return TRUE;

	CATCH_END(FALSE);
}

void init_peer_pg(pg_attribs *Attrib)
{
	memset(Attrib,0,sizeof(*Attrib));
	Attrib->protocol.RxFeatureTLVPresent =  FALSE;
}

boolean_t add_peer_pg(char *device_name)
{
	TRY_BEGIN;
	string dn(device_name); 

	pg_it it = peer_pg.find(dn);
	if (it == peer_pg.end()) {
		/* Device not present: add */
		pg_attribs *store = (pg_attribs*)malloc(sizeof(*store));

		if (!store) return FALSE;

		init_peer_pg( store);
		peer_pg.insert(make_pair(dn, store));
	}
	return TRUE;

	CATCH_END(FALSE);
}

void init_apptlv(app_attribs *Attrib, app_attribs *Store)
{
	memset(Store,0,sizeof(*Store));
	Attrib->protocol.Max_version     = DCB_APPTLV_MAX_VERSION;
	Attrib->protocol.State           = DCB_INIT;
	Attrib->protocol.Advertise_prev  = FALSE;
	Attrib->protocol.tlv_sent        = FALSE;
	memcpy(Store, Attrib, sizeof (*Attrib));
}

boolean_t add_apptlv(char *device_name, app_attribs *Attrib, u32 Subtype)
{
	/* Append Subtype to device name and then add.*/
	TRY_BEGIN;
	string dn(device_name); 

	char sTemp[SHORT_STRING];
	/* safe from overflow, only called from within for loop */
	snprintf(sTemp, SHORT_STRING, "%hu", Subtype);
	/* Append device name with subtype. */
	dn.append(sTemp);

	app_it it = apptlv.find(dn);
	if (it == apptlv.end()) {
		/* Device not present: add */
		app_attribs *store = (app_attribs*)malloc(sizeof(*store));

		if (!store) return FALSE;

		init_apptlv(Attrib, store);
		apptlv.insert(make_pair(dn, store));
	}
	else {  /* already in data store, just update it */
		(*it).second->protocol.Advertise_prev =
			(*it).second->protocol.Advertise;
		(*it).second->protocol.Advertise = Attrib->protocol.Advertise;
		(*it).second->protocol.Enable = Attrib->protocol.Enable;
		(*it).second->protocol.Willing = Attrib->protocol.Willing;
		(*it).second->Length = Attrib->Length;
		memcpy(&(*it).second->AppData, &(Attrib->AppData),
			Attrib->Length);
	}
	return TRUE;

	CATCH_END(FALSE);
}

void init_oper_apptlv(app_attribs *Store)
{
	memset(Store,0,sizeof(*Store));
}

boolean_t add_oper_apptlv(char *device_name, u32 Subtype)
{
	/* Append Subtype to device name and then add. */
	TRY_BEGIN;
	string dn(device_name); 
	char sTemp[SHORT_STRING];

	/* safe from overflow, only called from within for loop */
	snprintf(sTemp, SHORT_STRING, "%hu", Subtype);/* Localization OK */
	/* Append device name with subtype. */
	dn.append(sTemp);

	app_it it = oper_apptlv.find(dn);
	if (it == oper_apptlv.end()) {
		/* Device not present: add */
		app_attribs *store = (app_attribs*)malloc(sizeof(*store));
		if (!store) return FALSE;
		init_oper_apptlv(store);
		oper_apptlv.insert(make_pair(dn, store));
	}
	return TRUE;

	CATCH_END(FALSE);
}

void init_peer_apptlv(app_attribs *Store)
{
	memset(Store,0,sizeof(*Store));
}
boolean_t add_peer_apptlv(char *device_name, u32 Subtype)
{
	/* Append Subtype to device name and then add. */
	TRY_BEGIN;
	string dn(device_name); 
	char sTemp[SHORT_STRING];
	/* safe from overflow, only called from within for loop */
	snprintf(sTemp, SHORT_STRING, "%hu", Subtype);
	/* Append device name with subtype. */
	dn.append(sTemp);

	app_it it = peer_apptlv.find(dn);
	if (it == peer_apptlv.end()) {
		/* Device not present: add */
		app_attribs *store = (app_attribs*)malloc(sizeof(*store));
		if (!store) return FALSE;

		init_peer_apptlv(store);
		peer_apptlv.insert(make_pair(dn, store));
	}
	return TRUE;

	CATCH_END(FALSE);
}

void init_control_prot(control_protocol_attribs *Attrib)
{
	memset(Attrib,0,sizeof(*Attrib));
	Attrib->State = DCB_INIT;
	Attrib->Max_version = DCB_MAX_VERSION;
}

boolean_t add_control_protocol(char *device_name)
{
	TRY_BEGIN;
	string dn(device_name); 

	control_prot_it it = dcb_control_prot.find(dn);
	if (it == dcb_control_prot.end()) {
		/* Device not present: add */
		control_protocol_attribs *store =
			(control_protocol_attribs*)malloc(sizeof(*store));

		if (!store) return FALSE;

		init_control_prot( store);
		dcb_control_prot.insert(make_pair(dn, store));
	}
	return TRUE;

	CATCH_END(FALSE);
}

void init_peer_control_prot(control_protocol_attribs *Attrib)
{
	memset(Attrib,0,sizeof(*Attrib));
	Attrib->Oper_version = DCB_MAX_VERSION;
	Attrib->Max_version = DCB_MAX_VERSION;
	Attrib->RxDCBTLVState = DCB_PEER_NONE;
}

boolean_t add_peer_control_protocol(char *device_name)
{
	TRY_BEGIN;
	string dn(device_name); 

	control_prot_it it = dcb_peer_control_prot.find(dn);
	if (it == dcb_peer_control_prot.end()) {
		/* Device not present: add */
		control_protocol_attribs *store =
			(control_protocol_attribs*)malloc(sizeof(*store));

		if (!store) return FALSE;

		init_peer_control_prot( store);
		dcb_peer_control_prot.insert(make_pair(dn, store));
	}
	return TRUE;

	CATCH_END(FALSE);
}

void init_pfc(pfc_attribs *Attrib, pfc_attribs *Store)
{
	memset(Store,0,sizeof(*Store));
	Attrib->protocol.Max_version = DCB_PFC_MAX_VERSION;
	Attrib->protocol.State = DCB_INIT;
	Attrib->protocol.Advertise_prev = FALSE;
	Attrib->protocol.tlv_sent = FALSE;
	memcpy(Store, Attrib, sizeof(*Attrib));
}

boolean_t add_pfc(char *device_name, pfc_attribs *Attrib)
{
	TRY_BEGIN;
	string dn(device_name); 

	pfc_it it = pfc.find(dn);
	if (it == pfc.end()) {
		/* Device not present: add */
		pfc_attribs *store =
			(pfc_attribs*)malloc(sizeof(*store));

		if (!store) return FALSE;

		init_pfc(Attrib, store);
		pfc.insert(make_pair(dn, store));
	} else {  /* already present in store, just update */
		(*it).second->protocol.Advertise_prev =
			(*it).second->protocol.Advertise;
		(*it).second->protocol.Advertise = Attrib->protocol.Advertise;
		(*it).second->protocol.Enable = Attrib->protocol.Enable;
		(*it).second->protocol.Willing = Attrib->protocol.Willing;

		memcpy((*it).second->admin, &Attrib->admin,
			sizeof(Attrib->admin));
	}
	return TRUE;

	CATCH_END(FALSE);
}

void init_oper_pfc(pfc_attribs *Attrib)
{
	memset(Attrib,0,sizeof(*Attrib));
}

boolean_t add_oper_pfc(char *device_name)
{
	TRY_BEGIN;
	string dn(device_name); 

	pfc_it it = oper_pfc.find(dn);
	if (it == oper_pfc.end()) {
		/* Device not present: add */
		pfc_attribs *store =
			(pfc_attribs*)malloc(sizeof(*store));

		if (!store) return FALSE;

		init_oper_pfc(store);
		oper_pfc.insert(make_pair(dn, store));
	}
	return TRUE;

	CATCH_END(FALSE);
}

void init_peer_pfc(pfc_attribs *Attrib)
{
	memset(Attrib,0,sizeof(*Attrib));
	Attrib->protocol.RxFeatureTLVPresent =  FALSE;
}

boolean_t add_peer_pfc(char *device_name)
{
	TRY_BEGIN;
	string dn(device_name); 

	pfc_it it = peer_pfc.find(dn);
	if (it == peer_pfc.end()) {
		/* Device not present: add */
		pfc_attribs *store = (pfc_attribs*)malloc(sizeof(*store));

		if (!store) return FALSE;

		init_peer_pfc(store);
		peer_pfc.insert(make_pair(dn, store));
	}
	return TRUE;

	CATCH_END(FALSE);
}

void init_bwg_desc(bwg_descriptions *Attrib, bwg_descriptions *Store)
{
	memset(Store,0,sizeof(*Store));
	memcpy(Store, Attrib, sizeof(*Attrib));
}

boolean_t add_bwg_desc(char *device_name, bwg_descriptions *Attrib)
{
	TRY_BEGIN;
	string dn(device_name); 

	bwg_desc_it it = bwg_desc.find(dn);
	if (it == bwg_desc.end()) {
		/* Device not present: add */
		bwg_descriptions *store =
			(bwg_descriptions*)malloc(sizeof(*store));

		if (!store) return FALSE;

		init_bwg_desc(Attrib, store);
		store->valid_desc_id_count = 8;
		bwg_desc.insert(make_pair(dn, store));
	}
	return TRUE;

	CATCH_END(FALSE);
}

boolean_t add_adapter(char *device_name)
{
	u32 EventFlag = 0;
	full_dcb_attribs attribs;
	dcb_result sResult = dcb_success;

	printf("add_adapter (%s)\n", device_name);
	memset(&attribs, 0, sizeof(attribs));

	if ((sResult = get_persistent(device_name, &attribs)) != dcb_success) {
		printf("get_persistent returned error %d\n", sResult);
		log_message(MSG_ERR_ADD_CARD_FAILURE, "%s", device_name);
		return FALSE;
	}

	if ((sResult = dcb_check_config(&attribs.pg)) != dcb_success) {
		printf("Rule checker returned error %d\n", sResult);
	} else { /* big else */


		/* Create data store for the device. */
		if (!add_pg(device_name, &attribs.pg)) {
			log_message(MSG_ERR_RESOURCE_MEMORY, "%s", device_name);
			printf("add_pg error.\n");
			sResult = dcb_failed;
			goto add_adapter_error;
		}
		if (!add_pfc(device_name, &attribs.pfc)) {
			log_message(MSG_ERR_RESOURCE_MEMORY, "%s", device_name);
			printf("add_pfc error.\n");
			sResult = dcb_failed;
			goto add_adapter_error;
		}
		if (!add_bwg_desc(device_name, &attribs.descript)) {
			log_message(MSG_ERR_RESOURCE_MEMORY, "%s", device_name);
			printf("add_bwg_desc error.\n");
			sResult = dcb_failed;
			goto add_adapter_error;
		}
		if (!add_control_protocol(device_name)) {
			log_message(MSG_ERR_RESOURCE_MEMORY, "%s", device_name);
			printf("add_control_protocol error.\n");
			sResult = dcb_failed;
			goto add_adapter_error;
		}
		if (!add_peer_pg(device_name)) {
			log_message(MSG_ERR_RESOURCE_MEMORY, "%s", device_name);
			printf("add_peer_pg error.\n");
			sResult = dcb_failed;
			goto add_adapter_error;
		}
		if (!add_peer_pfc(device_name)) {
			log_message(MSG_ERR_RESOURCE_MEMORY, "%s", device_name);
			printf("add_peer_pfc error.\n");
			sResult = dcb_failed;
			goto add_adapter_error;
		}
		if (!add_peer_control_protocol(device_name)) {
			log_message(MSG_ERR_RESOURCE_MEMORY, "%s", device_name);
			printf("add_peer_control_protocol error.\n");
			sResult = dcb_failed;
			goto add_adapter_error;
		}
		if (!add_oper_pg(device_name)) {
			log_message(MSG_ERR_RESOURCE_MEMORY, "%s", device_name);
			printf("add_oper_pg error.\n");
			sResult = dcb_failed;
			goto add_adapter_error;
		}
		if (!add_oper_pfc(device_name)) {
			log_message(MSG_ERR_RESOURCE_MEMORY, "%s", device_name);
			printf("add_oper_pfc error.\n");
			sResult = dcb_failed;
			goto add_adapter_error;
		}
		/* Add APPTLV for supported Subtypes. */
		for(int i = 0; i < DCB_MAX_APPTLV ; i++) {
			if (!add_apptlv(device_name,
				&attribs.app[i], i)) {
				log_message(MSG_ERR_RESOURCE_MEMORY, "%s", device_name);
				printf("add_apptlv error.\n");
				sResult = dcb_failed;
				goto add_adapter_error;
			}
			if (!add_oper_apptlv(device_name, i)) {
				log_message(MSG_ERR_RESOURCE_MEMORY, "%s", device_name);
				printf("add_oper_apptlv error.\n");
				sResult = dcb_failed;
				goto add_adapter_error;
			}
			if (!add_peer_apptlv(device_name, i)) {
				log_message(MSG_ERR_RESOURCE_MEMORY, "%s", device_name);
				printf("add_peer_apptlv error.\n");
				sResult = dcb_failed;
				goto add_adapter_error;
			}
		}

		/* Initialize features state machines for PG and PFC and
		 * APPTLVs. */
		if (run_feature_protocol(device_name,
			DCB_LOCAL_CHANGE_PG) !=
			dcb_success) {
			printf("run_feature_protocol error.\n");
			sResult = dcb_failed;
			goto add_adapter_error;
		}
		if (run_feature_protocol(device_name,
			DCB_LOCAL_CHANGE_PFC) != dcb_success) {
			printf("run_feature_protocol error.\n");
			sResult = dcb_failed;
			goto add_adapter_error;
		}
		/* If APPTLV subtypes are supported then run the
		 * feature_protocol for those supported subtypes. */
		for(int i = 0; i < DCB_MAX_APPTLV ; i++) {
			if (run_feature_protocol(device_name,
				DCB_LOCAL_CHANGE_APPTLV, i) !=
				dcb_success) {
				printf("run_feature_protocol error.\n");
				sResult = dcb_failed;
				goto add_adapter_error;
			}
		}
		/* apply all feature setting to the driver:linux only */
		set_hw_all(device_name);

		EventFlag = 0;
		DCB_SET_FLAGS(EventFlag, DCB_LOCAL_CHANGE_PG |
			DCB_LOCAL_CHANGE_PFC| DCB_LOCAL_CHANGE_APPTLV);
		/* Initialize control state machine */
		if (run_control_protocol (device_name, EventFlag) !=
			dcb_success) {
			printf("run_control_protocol error.\n");
			sResult = dcb_failed;
			goto add_adapter_error;
		}
	}	/* big else on top */

add_adapter_error:
	if (sResult != dcb_success) {
		printf("add_adapter: Service unable to use network adapter.\n");
		log_message(MSG_ERR_ADD_CARD_FAILURE, "%s", device_name);
		return FALSE;
	}

	if (add_port(device_name) < 0) {
		printf("add_adapter: add_port failed.\n");
		log_message(MSG_ERR_ADD_CARD_FAILURE, "%s", device_name);
		return FALSE;
	}
	return TRUE;
}

boolean_t remove_adapter(char *device_name)
{
	/* Set feature state machines to DCB_CLOSED first */

	/* NOTE: the parameter "device_name" is freed during this routine
	 * and can not be used */
	char devName[MAX_DEVICE_NAME_LEN]; /* local copy of device_name */
	feature_protocol_attribs *pFP;
	void  *itp = NULL;

	/* save device name for remove port */
	if (device_name == NULL) {
		printf("remove_adapter: null device\n");
		return FALSE;
	}
	strncpy (devName, device_name, MAX_DEVICE_NAME_LEN);

	TRY_BEGIN;
	string dn(device_name); 
	pg_it itpg = pg.find(dn);
	if (itpg != pg.end()) {
		/* save the reserved memory here so we can free
		 * it since erase wipes out the it structure 
		 * (but does not free the memory) */
		itp = (void *) itpg->second;

		/* Get the feature protocol for local store */
		pFP = (&((*itpg).second)->protocol);
		/* Set feature protocol state to closed. */
		pFP->State = DCB_CLOSED;
		/* this line frees the param of this function! */
		pg.erase(dn);
		if (itp) free (itp);
		itp = NULL;
	} else {
		printf("remove_adapter: pg not found\n");
	}

		pfc_it itpfc = pfc.find(dn);
		if (itpfc != pfc.end()) {
		itp = (void *) itpfc->second;
		pFP = (&((*itpfc).second)->protocol);
		pFP->State = DCB_CLOSED;
		pfc.erase(dn);
		if (itp)
			free(itp);
		itp = NULL;
	} else {
		printf("remove_adapter: pfc not found\n");
	}

	bwg_desc_it itbwg = bwg_desc.find(dn);
	if (itbwg != bwg_desc.end()) {
		itp = (void *) itbwg->second;
		bwg_desc.erase(dn);
		if (itp)
			free(itp);
		itp = NULL;
	} else {
		printf("remove_adapter: bwg not found\n");
	}

	control_prot_it itcp = dcb_control_prot.find(dn);
	if (itcp != dcb_control_prot.end()) {
		itp = (void *) itcp->second;
		dcb_control_prot.erase(dn);
		if (itp)
			free (itp);
		itp = NULL;
	} else {
		printf("remove_adapter: ctrl not found\n");
	}

	pg_it itprpg = peer_pg.find(dn);
	if (itprpg != peer_pg.end()) {
		itp = (void *) itprpg->second;
		pFP = (&((*itprpg).second)->protocol);
		pFP->State = DCB_CLOSED;
		peer_pg.erase(dn);
		if (itp) free (itp);
		itp = NULL;
	} else {
		printf("remove_adapter: peer pg not found\n");
	}

	pfc_it itprpfc = peer_pfc.find(dn);
	if (itprpfc != peer_pfc.end()) {
		itp = (void *) itprpfc->second;
		pFP = (&((*itprpfc).second)->protocol);
		pFP->State = DCB_CLOSED;
		peer_pfc.erase(dn);
		if (itp)
			free (itp);
		itp = NULL;
	} else {
		printf("remove_adapter: peer pfc not found\n");
	}

	control_prot_it itpcp = dcb_peer_control_prot.find(dn);
	if (itpcp != dcb_peer_control_prot.end()) {
		itp = (void *) itpcp->second;
		dcb_peer_control_prot.erase(dn);
		if (itp)
			free (itp);
		itp = NULL;
	} else {
		printf("remove_adapter: peer ctrl not found\n");
	}

	pg_it itopg = oper_pg.find(dn);
	if (itopg != oper_pg.end()) {
		itp = (void *) itopg->second;
		pFP = (&((*itopg).second)->protocol);
		pFP->State = DCB_CLOSED;
		oper_pg.erase(dn);
		if (itp)
			free(itp);
		itp = NULL;
	} else {
		printf("remove_adapter: oper pg not found\n");
	}

	pfc_it itopfc = oper_pfc.find(dn);
	if (itopfc != oper_pfc.end()) {
		itp = (void *) itopfc->second;
		pFP = (&((*itopfc).second)->protocol);
		pFP->State = DCB_CLOSED;
		oper_pfc.erase(dn);
		if (itp) free (itp);
		itp = NULL;
	} else {
		printf("remove_adapter: oper pfc not found\n");
	}

	/* Get the APP TLV  and erase. */
	for(int i = 0; i < DCB_MAX_APPTLV ; i++) {
		string dn1 = dn;
		char sTemp[SHORT_STRING];
		/* Localization OK */
		snprintf(sTemp, SHORT_STRING, "%hu", i);
		/* Append device name with subtype */
		dn1.append(sTemp);
		app_it itapp = apptlv.find(dn1);
		if (itapp != apptlv.end()) {
			itp = (void *) itapp->second;
			apptlv.erase(dn1);
			if (itp) free (itp);
			itp = NULL;
		}
		itapp = oper_apptlv.find(dn1);
		if (itapp != oper_apptlv.end()) {
			itp = (void *) itapp->second;
			oper_apptlv.erase(dn1);
			if (itp) free (itp);
			itp = NULL;
		}
		itapp = peer_apptlv.find(dn1);
		if (itapp != peer_apptlv.end()) {
			itp = (void *) itapp->second;
			peer_apptlv.erase(dn1);
			if (itp) free (itp);
			itp = NULL;
		}
	}

	CATCH
	{
		return FALSE;
	}


	if (remove_port(devName) < 0) {
		printf("remove_port: failed\n");
		return FALSE;
	}
	return TRUE;
}

/* Function to find all the dcb devices from store and remove them. */
void remove_all_adapters()
{
	pg_it it;
	char sTmp[MAX_DEVICE_NAME_LEN*2];

	it = pg.begin();
	while (it != pg.end()) {
		/* prepare sTmp in case of error */
		snprintf(sTmp, MAX_DEVICE_NAME_LEN*2, /* Localization OK */
			"Remove_all_adapters error: Bad device name: %.*s\n",
			MAX_DEVICE_NAME_LEN, it->first.c_str());
		if (remove_adapter((char *) (it++->first.c_str())) == FALSE) {
			printf(sTmp);
		}
	}

	return;
}

boolean_t add_pg_defaults()
{
	pg_attribs  pg_data;
	char        sTmp[MAX_DESCRIPTION_LEN];
	boolean_t   result = TRUE;
	int         index, portion, rmndr, temp;

	memset(&pg_data, 0, sizeof(pg_attribs));

	pg_data.protocol.Enable = 1;
	pg_data.protocol.Willing = 1;
	if (bLLDPPresent)
		pg_data.protocol.Advertise = 1;
	else
		pg_data.protocol.Advertise = 0;

	portion = BW_PERCENT/MAX_BANDWIDTH_GROUPS;
	rmndr   = BW_PERCENT % MAX_BANDWIDTH_GROUPS;

	temp = rmndr;
	for (index=0; index < MAX_BANDWIDTH_GROUPS; index++) {
		pg_data.tx.bwg_percent_of_link[index] = (u8)portion;
		if (temp >0) {
			pg_data.tx.bwg_percent_of_link[index] += 1;
			temp--;
		}
	}
	for (index=0; index < MAX_USER_PRIORITIES; index++) {
		pg_data.tx.up[index].bandwidth_group_mapping = (u8)index;
		pg_data.tx.up[index].percent_of_bandwidth_group_cap =
			BW_PERCENT;
		pg_data.tx.up[index].strict_priority = dcb_none;
		pg_data.tx.up[index].traffic_class_mapping = (u8)(index);
	}
	temp = rmndr;
	for (index=0; index < MAX_BANDWIDTH_GROUPS; index++) {
		pg_data.rx.bwg_percent_of_link[index] = (u8)portion;
		if (temp >0) {
			pg_data.rx.bwg_percent_of_link[index]++;
			temp--;
		}
	}
	for (index=0; index < MAX_USER_PRIORITIES; index++) {
		pg_data.rx.up[index].bandwidth_group_mapping = (u8)index;
		pg_data.rx.up[index].percent_of_bandwidth_group_cap =
			BW_PERCENT;
		pg_data.rx.up[index].strict_priority = dcb_none;
		pg_data.rx.up[index].traffic_class_mapping = (u8)(index);
	}

	snprintf(sTmp, MAX_DESCRIPTION_LEN, DEF_CFG_STORE);

	/* Create pg default data store for the device. */
	if (!add_pg(sTmp, &pg_data))
		result = FALSE;

	return result;
}

boolean_t remove_pg_defaults()
{
	char   sTmp[MAX_DESCRIPTION_LEN];
	void * itp = NULL;

	snprintf(sTmp, MAX_DESCRIPTION_LEN, DEF_CFG_STORE);
	pg_it itpg = pg.find(sTmp);
	if (itpg == pg.end())
		return FALSE;
	itp = (void *) itpg->second;
	/* erase and free memory */
	pg.erase(sTmp);
	if (itp) {
		free (itp);
		itp = NULL;
	}
	return TRUE;
}

boolean_t add_pfc_defaults()
{
	pfc_attribs pfc_data;
	char        sTmp[MAX_DESCRIPTION_LEN];
	boolean_t   result = TRUE;
	int         index;

	memset (&pfc_data, 0, sizeof(pfc_attribs));

	pfc_data.protocol.Enable = 1;
	pfc_data.protocol.Willing = 1;
	if (bLLDPPresent)
		pfc_data.protocol.Advertise = 1;
	else
		pfc_data.protocol.Advertise = 0;

	for (index=0; index < MAX_TRAFFIC_CLASSES; index++) {
		pfc_data.admin[index] = pfc_disabled;
	}

	snprintf(sTmp, MAX_DESCRIPTION_LEN, DEF_CFG_STORE);
	/* Create pfc default data store for the device. */
	if (!add_pfc(sTmp, &pfc_data))
		result = FALSE;

	return result;
}

boolean_t remove_pfc_defaults()
{
	char   sTmp[MAX_DESCRIPTION_LEN];
	void * itp = NULL;

	snprintf(sTmp, MAX_DESCRIPTION_LEN, DEF_CFG_STORE);
	pfc_it itpfc = pfc.find(sTmp);
	if (itpfc == pfc.end()) {
		return FALSE;
	}
	itp = (void *) itpfc->second;
	/* erase and free memory */
	pfc.erase(sTmp);
	if (itp) {
		free (itp);
		itp = NULL;
	}
	return TRUE;
}

boolean_t add_app_defaults(u32 subtype)
{
	app_attribs app_data;
	char        sTmp[MAX_DESCRIPTION_LEN];
	boolean_t   result = TRUE;

	memset (&app_data, 0, sizeof(app_attribs));

	app_data.protocol.Enable = 0;
	app_data.protocol.Willing = 1;
	if (bLLDPPresent)
		app_data.protocol.Advertise = 1;
	else
		app_data.protocol.Advertise = 0;

	snprintf(sTmp, MAX_DESCRIPTION_LEN, DEF_CFG_STORE);  /* Localization OK */

	switch (subtype) {
	case 0:  /* FCoE subtype */
		app_data.Length = 1;
		app_data.AppData[0] = 0x08;
		break;
	default:
		break;
	}

	/* Create app default data store for the device and app subtype. */
	if (!add_apptlv(sTmp, &app_data, subtype))
		result = FALSE;

	return result;
}

boolean_t remove_app_defaults(u32 subtype)
{
	char   sTmp[MAX_DESCRIPTION_LEN];
	void * itp = NULL;

	snprintf(sTmp, MAX_DESCRIPTION_LEN, "%s%hu",   /* Localization OK */
			 DEF_CFG_STORE, subtype);
	app_it it = apptlv.find(sTmp);

	if (it == apptlv.end()) {
		return FALSE;
	}
	itp = (void *) it->second;
	/* erase and free memory */
	apptlv.erase(sTmp);
	if (itp) {
		free (itp);
		itp = NULL;
	}
	return TRUE;
}

dcb_result get_pg(char *device_name,  pg_attribs *pg_data)
{
	TRY_BEGIN;
	string dn(device_name); 

	dcb_result result = dcb_success;
	full_dcb_attribs attribs;

	memset(&attribs, 0, sizeof(attribs));
	if (!pg_data) return dcb_bad_params;
	pg_it it = pg.find(dn);
	if (it != pg.end()) {
		memcpy( pg_data, (*it).second, sizeof(*pg_data));
	}
	else {
		result = get_persistent(device_name, &attribs);
		if (result == dcb_success)
			memcpy( pg_data, &attribs.pg, sizeof(*pg_data));
		else
			result = dcb_device_not_found;
	}
	return result;

	CATCH_END(dcb_failed);
}

dcb_result get_pg_dstore(char *device_name,  pg_attribs *pg_data)
{
	TRY_BEGIN;
	string dn(device_name); 

	dcb_result result = dcb_success;

	if (!pg_data) return dcb_bad_params;
	pg_it it = pg.find(dn);
	if (it != pg.end()) {
		memcpy( pg_data, (*it).second, sizeof(*pg_data));
	} else {
		result = dcb_device_not_found;
	}
		return result;

	CATCH_END(dcb_failed);
}

dcb_result get_oper_pg(char *device_name,  pg_attribs *pg_data)
{
	TRY_BEGIN;
	string dn(device_name); 

	dcb_result result = dcb_success;

	if (!pg_data) return dcb_bad_params;
	pg_it it = oper_pg.find(dn);
	if (it != oper_pg.end()) {
		memcpy(pg_data, (*it).second, sizeof(*pg_data));
	} else {
		result = dcb_device_not_found;
	}
	return result;

	CATCH_END(dcb_failed);
}

dcb_result get_peer_pg(char *device_name,  pg_attribs *pg_data)
{
	TRY_BEGIN;
	string dn(device_name); 

	dcb_result result = dcb_success;

	if (!pg_data) return dcb_bad_params;
	pg_it it = peer_pg.find(dn);
	if (it != peer_pg.end()) {
		memcpy(pg_data, (*it).second, sizeof(*pg_data));
	} else {
		result = dcb_device_not_found;
	}

	return result;

	CATCH_END(dcb_failed);
}

void mark_pg_sent(char *device_name)
{
	TRY_BEGIN;
	string dn(device_name); 

	pg_it it = pg.find(dn);
	if (it != pg.end()) {
		(*it).second->protocol.tlv_sent = TRUE;
	}
	CATCH_END_VOID;
}

void mark_pfc_sent(char *device_name)
{
	TRY_BEGIN;
	string dn(device_name); 

	pfc_it it = pfc.find(dn);
	if (it != pfc.end()) {
		(*it).second->protocol.tlv_sent = TRUE;
	}
	CATCH_END_VOID;
}

void mark_app_sent(char *device_name, u32 subtype)
{
	TRY_BEGIN;
	string dn(device_name); 

	char stmp[SHORT_STRING];
	snprintf(stmp, SHORT_STRING, "%hu", subtype);	  /* Localization OK */

	/* append device name with subtype */
	dn.append(stmp);

	app_it it = apptlv.find(dn);
	if (it != apptlv.end()) {
		(*it).second->protocol.tlv_sent = TRUE;
	}
	CATCH_END_VOID;
}

dcb_result put_pg(char *device_name,  pg_attribs *pg_data)
{
	TRY_BEGIN;
	u32              EventFlag = 0;
	dcb_result       result = dcb_success;
	full_dcb_attribs attribs;
	boolean_t        bChange = FALSE;
	string           dn(device_name);

	if (!pg_data) return dcb_bad_params;

	memset(&attribs, 0, sizeof(attribs));
	pg_it it = pg.find(dn);
	if (it != pg.end()) {
		/* Lock the data first */

		/* detect no config change */
		if (memcmp((*it).second, pg_data, sizeof(*pg_data)) == 0) {
			goto Exit;
		}

		/* Check the rules */
		if (dcb_check_config(pg_data) != dcb_success) {
			printf("Rule checking failed in put_pg()\n");
			goto ErrNoDevice;
		}
		bChange = TRUE;
		if (set_persistent(device_name, pg_data,
			NULL, NULL,NULL,0) !=
			dcb_success) {
			printf("Set persistent failed put_pg()\n");
			goto ErrNoDevice;
		}
		/* Copy the writtable protocol * variables */
		feature_protocol_attribs *dStore = &((*it).second->protocol);
		dStore->Advertise_prev  = dStore->Advertise;
		dStore->Advertise       = pg_data->protocol.Advertise;
		dStore->Enable          = pg_data->protocol.Enable;
		dStore->Willing         = pg_data->protocol.Willing;
		dStore->tlv_sent        = FALSE;

		memcpy(&(*it).second->rx, &(pg_data->rx), sizeof(pg_data->rx));
		memcpy(&(*it).second->tx, &(pg_data->tx), sizeof(pg_data->tx));

		DCB_SET_FLAGS(EventFlag, DCB_LOCAL_CHANGE_PG);

		/* Run the protocol */
		if (bChange) {
			result = run_dcb_protocol(device_name, EventFlag);
		}
	} else {
		/* Not in DCB data store, so store in persistent storage */
		if (get_persistent(device_name, &attribs) == dcb_success) {
			if (set_persistent(device_name, pg_data, NULL,
				&attribs.descript, NULL, 0) != dcb_success) {
				printf("Set persistent failed in put_pg()\n");
				result = dcb_device_not_found;
			}
		} else {
			result = dcb_device_not_found;
		}
	}
Exit:
	return result;

ErrNoDevice:
	return dcb_device_not_found;

	CATCH_END(dcb_failed);
}

dcb_result put_peer_pg(char *device_name,  pg_attribs *peer_pg_data)
{
	/* this function relies on the caller to acquire the DCB lock */
	TRY_BEGIN;
	string dn(device_name); 
	dcb_result result = dcb_success;
	feature_protocol_attribs *dStore;

	if (!peer_pg_data) return dcb_bad_params;
	pg_it peer_it = peer_pg.find(dn);
	if (peer_it == peer_pg.end()) {
		printf("could not find peer_pg data for %s\n", device_name);
		result = dcb_device_not_found;
		goto Exit;
	}

	/* detect config change */
	if (memcmp((*peer_it).second, peer_pg_data,
		sizeof(*peer_pg_data)) == 0) {
		goto Exit;
	}

	/* Copy the writtable protocol variables. */
	dStore = &((*peer_it).second->protocol);
	dStore->Advertise_prev = dStore->Advertise;
	dStore->Advertise = peer_pg_data->protocol.Advertise;
	dStore->Enable = peer_pg_data->protocol.Enable;
	dStore->Willing = peer_pg_data->protocol.Willing;
	dStore->Oper_version = peer_pg_data->protocol.Oper_version;
	dStore->Max_version = peer_pg_data->protocol.Max_version;
	dStore->RxFeatureTLVPresent =
		peer_pg_data->protocol.RxFeatureTLVPresent;
	dStore->Error = peer_pg_data->protocol.Error;

	memcpy(&(*peer_it).second->rx, &(peer_pg_data->rx),
		sizeof(peer_pg_data->rx));
	memcpy(&(*peer_it).second->tx, &(peer_pg_data->tx),
		sizeof(peer_pg_data->tx));

Exit:
	return result;

	CATCH_END(dcb_failed);
}


dcb_result get_pfc(char *device_name, pfc_attribs *pfc_data)
{
	TRY_BEGIN;
	string dn(device_name); 

	dcb_result result = dcb_success;
	full_dcb_attribs attribs;

	memset(&attribs, 0, sizeof(attribs));
	if (!pfc_data)
		return dcb_bad_params;
	pfc_it it = pfc.find(dn);
	if (it != pfc.end()) {
		memcpy( pfc_data, (*it).second, sizeof(*pfc_data));
	} else {
		result = get_persistent(device_name, &attribs);
		if (result == dcb_success)
			memcpy( pfc_data, &attribs.pfc, sizeof(*pfc_data));
		else
			result = dcb_device_not_found;
	}
	return result;

	CATCH_END(dcb_failed);
}

dcb_result get_oper_pfc(char *device_name, pfc_attribs *pfc_data)
{
	TRY_BEGIN;
	string dn(device_name); 

	dcb_result result = dcb_success;

	if (!pfc_data) return dcb_bad_params;
	pfc_it it = oper_pfc.find(dn);
	if (it != oper_pfc.end()) {
		memcpy( pfc_data, (*it).second, sizeof(*pfc_data));
	} else {
		result = dcb_device_not_found;
	}
	return result;    

	CATCH_END(dcb_failed);
}

dcb_result get_peer_pfc(char *device_name, pfc_attribs *pfc_data)
{
	TRY_BEGIN;
	string dn(device_name); 

	dcb_result result = dcb_success;

	if (!pfc_data) return dcb_bad_params;
	pfc_it it = peer_pfc.find(dn);
	if (it != peer_pfc.end()) {
		memcpy( pfc_data, (*it).second, sizeof(*pfc_data));
	} else {
		result = dcb_device_not_found;
	}
	return result;

	CATCH_END(dcb_failed);
}

dcb_result put_pfc(char *device_name, pfc_attribs *pfc_data)
{
	TRY_BEGIN;
	u32              EventFlag = 0;
	dcb_result       result = dcb_success;
	full_dcb_attribs attribs;
	boolean_t        bChange = FALSE;
	string           dn(device_name); 

	if (!pfc_data) return dcb_bad_params;

	memset(&attribs, 0, sizeof(attribs));
	pfc_it it = pfc.find(dn);
	if (it != pfc.end()) {
		/* Lock the data first */

		/* detect no config change */
		if (memcmp((*it).second, pfc_data, sizeof(*pfc_data)) == 0) {
			goto Exit;
		}
		bChange = TRUE;
		if (set_persistent(device_name, NULL, pfc_data, NULL,NULL,0) 
				!= dcb_success) {
			printf("Set persistent failed in put_pfc()\n");
			result = dcb_device_not_found;
			goto Exit;
			}
		feature_protocol_attribs *dStore = &((*it).second->protocol);
		dStore->Advertise_prev  = dStore->Advertise;
		dStore->Advertise       = pfc_data->protocol.Advertise;
		dStore->Enable          = pfc_data->protocol.Enable;
		dStore->Willing         = pfc_data->protocol.Willing;
		dStore->tlv_sent        = FALSE;

		memcpy((*it).second->admin, pfc_data->admin,
			sizeof(pfc_data->admin));

		/* Run the protocol */
		DCB_SET_FLAGS(EventFlag, DCB_LOCAL_CHANGE_PFC);
		if (bChange) {
			result = run_dcb_protocol(device_name, EventFlag);
		}
	} else {
		/* Store in persistent storage - though not in DCB data store */
		if (get_persistent(device_name, &attribs) == dcb_success) {
			if (set_persistent(device_name, NULL, pfc_data,
				&attribs.descript, NULL, 0) != dcb_success) {
				result = dcb_device_not_found;
			}
		}
		else
			result = dcb_device_not_found;
	}
Exit:
	return result;

	CATCH_END(dcb_failed);
}

dcb_result put_peer_pfc(char *device_name, pfc_attribs *peer_pfc_data)
{
	/* this function relies on the caller to acquire the DCB lock */
	TRY_BEGIN;
	string dn(device_name); 
	dcb_result result = dcb_success;
	feature_protocol_attribs *dStore;

	if (!peer_pfc_data) return dcb_bad_params;
	pfc_it peer_it = peer_pfc.find(dn);
	if (peer_it == peer_pfc.end()){
		printf("putting peer_pfc data - bad device name\n");
		result = dcb_device_not_found;
		goto Exit;
	}
	/* detect config change */
	if (memcmp((*peer_it).second, peer_pfc_data,
		sizeof(*peer_pfc_data)) == 0) {
		goto Exit;
	}

	dStore = &((*peer_it).second->protocol);
	dStore->Advertise_prev  = dStore->Advertise;
	dStore->Advertise       = peer_pfc_data->protocol.Advertise;
	dStore->Enable          = peer_pfc_data->protocol.Enable;
	dStore->Willing         = peer_pfc_data->protocol.Willing;
	dStore->Oper_version    = peer_pfc_data->protocol.Oper_version;
	dStore->Max_version     = peer_pfc_data->protocol.Max_version;
	dStore->RxFeatureTLVPresent = 
		peer_pfc_data->protocol.RxFeatureTLVPresent;
	dStore->Error           = peer_pfc_data->protocol.Error;

	memcpy((*peer_it).second->admin, &peer_pfc_data->admin,
		sizeof(peer_pfc_data->admin));
Exit:
	return result;

	CATCH_END(dcb_failed);
}

dcb_result get_app(char *device_name, u32 subtype, app_attribs *app_data)
{
	TRY_BEGIN;
	string dn(device_name); 

	dcb_result result = dcb_success;
	full_dcb_attribs attribs;
	char stmp[SHORT_STRING];

	memset(&attribs, 0, sizeof(attribs));
	/* append device name with subtype */
	snprintf(stmp, SHORT_STRING, "%hu", subtype);	  /* Localization OK */
	dn.append(stmp);

	app_it it = apptlv.find(dn);

	if (it != apptlv.end()) {
		memcpy(app_data, (*it).second, sizeof(*app_data));
	} else {
		result = get_persistent(device_name, &attribs);
		if (result == dcb_success) {
			memcpy( app_data, &attribs.app[subtype],
			sizeof(*app_data));
		}
		else
			result = dcb_device_not_found;
	}
	return result;

	CATCH_END(dcb_failed);
}

dcb_result get_oper_app(char *device_name, u32 subtype, app_attribs *app_data)
{
	TRY_BEGIN;
	string dn(device_name); 

	dcb_result result = dcb_success;
	char stmp[SHORT_STRING];
	snprintf(stmp, SHORT_STRING, "%hu", subtype);	  /* Localization OK */

	/* append device name with subtype */
	dn.append(stmp);

	app_it it = oper_apptlv.find(dn);
	if (it != oper_apptlv.end()) {
		memcpy(app_data, (*it).second, sizeof(*app_data));
	} else {
		result = dcb_device_not_found;
	}
	return result;

	CATCH_END(dcb_failed);
}

dcb_result get_peer_app(char *device_name, u32 subtype, app_attribs *app_data)
{
	TRY_BEGIN;
	string dn(device_name); 
	dcb_result result = dcb_success;
	char stmp[SHORT_STRING];
	snprintf(stmp, SHORT_STRING, "%hu", subtype);	  /* Localization OK */

	/* append device name with subtype */
	dn.append(stmp);

	app_it it = peer_apptlv.find(dn);
	if (it != peer_apptlv.end()) {
		memcpy(app_data, (*it).second, sizeof(*app_data));
	} else {
		result = dcb_device_not_found;
	}
	return result;

	CATCH_END(dcb_failed);
}

dcb_result put_app(char *device_name, u32 subtype, app_attribs *app_data)
{
	TRY_BEGIN;
	dcb_result       result = dcb_success;
	u32              EventFlag = 0;
	char             stmp[SHORT_STRING];
	full_dcb_attribs attribs;
	boolean_t        bChange = FALSE;
	string           dn(device_name); 

	/* append device name with subtype */
	snprintf(stmp, SHORT_STRING, "%hu", subtype); /* Localization OK */
	dn.append(stmp);

	if (!app_data) return dcb_bad_params;

	memset(&attribs, 0, sizeof(attribs));
	app_it it = apptlv.find(dn);
	if (it != apptlv.end()) {
		/* Lock the data first */

		/* detect no config change */
		if (memcmp((*it).second, app_data,
			sizeof(*app_data)) == 0) {
			goto Exit;
		}
		/* Store in persistent storage */
		bChange = TRUE;
		if (set_persistent(device_name, NULL,
			NULL, NULL, app_data,
			(u8)subtype) != dcb_success) {
			printf("Set persistent failed in put_app()\n");
			return dcb_device_not_found;
		}
		feature_protocol_attribs *dStore = &((*it).second->protocol);
		dStore->Advertise_prev  = dStore->Advertise;
		dStore->Advertise       = app_data->protocol.Advertise;
		dStore->Enable          = app_data->protocol.Enable;
		dStore->Willing         = app_data->protocol.Willing;
		dStore->tlv_sent        = FALSE;
		if (app_data->Length) {
			(*it).second->Length = app_data->Length;
			memcpy(&(*it).second->AppData, &(app_data->AppData),
				app_data->Length);
		}
		DCB_SET_FLAGS(EventFlag, DCB_LOCAL_CHANGE_APPTLV);

		if (bChange) {
			result = run_dcb_protocol (device_name, 
				EventFlag, subtype);
		}

	} else {
		/* Not in DCB data store, store in persistent storage */
		if (get_persistent(device_name, &attribs) == dcb_success) {
			if (set_persistent(device_name, NULL, NULL,
				&attribs.descript, app_data, (u8)subtype) !=
				dcb_success) {
				printf("Set persistent failed in put_app()\n");
				result = dcb_device_not_found;
			}
		} else {
			result = dcb_device_not_found;
		}
	}
Exit:
	return result;

	CATCH_END(dcb_failed);
}

dcb_result put_peer_app(char *device_name, u32 subtype,
			app_attribs *peer_app_data)
{
	/* this function relies on the caller to acquire the DCB lock */
	TRY_BEGIN;
	string dn(device_name); 
	dcb_result result = dcb_success;
	char stmp[SHORT_STRING];
	feature_protocol_attribs *dStore;

	/* append device name with subtype */
	snprintf(stmp, SHORT_STRING, "%hu", subtype); /* Localization OK */
	dn.append(stmp);

	if (!peer_app_data) return dcb_bad_params;
	app_it peer_it = peer_apptlv.find(dn);
	if (peer_it == peer_apptlv.end()) {
		printf("putting peer_app data - bad device name\n");
		result = dcb_device_not_found;
		goto Exit;
	}
	if (memcmp((*peer_it).second, peer_app_data,
		sizeof(*peer_app_data)) == 0) {
		goto Exit;
	}

	dStore = &((*peer_it).second->protocol);
	dStore->Advertise_prev = dStore->Advertise;
	dStore->Advertise      = peer_app_data->protocol.Advertise;
	dStore->Enable         = peer_app_data->protocol.Enable;
	dStore->Willing        = peer_app_data->protocol.Willing;
	dStore->Oper_version   = peer_app_data->protocol.Oper_version;
	dStore->Max_version    = peer_app_data->protocol.Max_version;
	dStore->RxFeatureTLVPresent =
		peer_app_data->protocol.RxFeatureTLVPresent;
	dStore->Error = peer_app_data->protocol.Error;
	(*peer_it).second->Length = peer_app_data->Length;
	memcpy(&(*peer_it).second->AppData, &(peer_app_data->AppData),
		peer_app_data->Length);

Exit:
	return result;

	CATCH_END(dcb_failed);
}

dcb_result get_control(char *device_name,
			control_protocol_attribs *control_data)
{
	TRY_BEGIN;
	string dn(device_name); 
	dcb_result result = dcb_success;

	if (!control_data) return dcb_bad_params;
	control_prot_it it = dcb_control_prot.find(dn);
	if (it != dcb_control_prot.end()) {
		memcpy( control_data, (*it).second, sizeof(*control_data));
	} else {
		result = dcb_device_not_found;
	}
	return result;

	CATCH_END(dcb_failed);
}

dcb_result get_peer_control(char *device_name,
				control_protocol_attribs *peer_control_data)
{
	TRY_BEGIN;
	string dn(device_name); 

	dcb_result result = dcb_success;

	if (!peer_control_data) return dcb_bad_params;
	control_prot_it it = dcb_peer_control_prot.find(dn);
	if (it != dcb_peer_control_prot.end()) {
		memcpy(peer_control_data, (*it).second, sizeof(*peer_control_data));
	} else {
		result = dcb_device_not_found;
	}
	return result;

	CATCH_END(dcb_failed);
}


dcb_result put_peer_control(char *device_name,
				control_protocol_attribs *peer_control_data)
{
	/* this function relies on the caller to acquire the DCB lock */
	TRY_BEGIN;
	string dn(device_name); 
	dcb_result result = dcb_success;

	if (!peer_control_data) return dcb_bad_params;
	control_prot_it peer_control_protocol = dcb_peer_control_prot.find(dn);

	if (peer_control_protocol != dcb_peer_control_prot.end()) {
		peer_control_protocol->second->SeqNo =
			peer_control_data->SeqNo;
		peer_control_protocol->second->AckNo =
			peer_control_data->AckNo;
		peer_control_protocol->second->Max_version =
			peer_control_data->Max_version;
		peer_control_protocol->second->Oper_version =
			peer_control_data->Oper_version;
		peer_control_protocol->second->RxDCBTLVState =
			peer_control_data->RxDCBTLVState;
	} else {
		result = dcb_device_not_found;
	}
	return result;

	CATCH_END(dcb_failed);
}

dcb_result get_bwg_descrpt(char *device_name, u8 bwgid, char **name)
{
	TRY_BEGIN;
	string dn(device_name); 
	dcb_result result = dcb_success;
	full_dcb_attribs attribs;
	int size;

	if (*name != NULL)
		free(*name);

	memset(&attribs, 0, sizeof(attribs));
	bwg_desc_it it = bwg_desc.find(dn);

	if ((it != bwg_desc.end()) &&
		(bwgid < ((*it).second)->valid_desc_id_count)) {
		size = (int)strlen(((*it).second)->bwg_desc_of_id[bwgid]) +
			sizeof(char);  /* Localization OK */
		*name = (char*)malloc(size);
		if (*name != NULL) {
			strncpy(*name, ((*it).second)->bwg_desc_of_id[bwgid],
					size); /* Localization OK */
		} else {
			goto Error;
		}
	} else {
		result = get_persistent(device_name, &attribs);
		if (result == dcb_success) {
			size = (int)strlen(
				attribs.descript.bwg_desc_of_id[bwgid]) +
				sizeof(char);
			*name = (char*)malloc(size);
			if (*name != NULL) {
				strncpy(*name,
					attribs.descript.bwg_desc_of_id[bwgid],
					size); /* Localization OK */
			} else {
				goto Error;
			}
		} else {
			result = dcb_device_not_found;
		}
	}
	return result;

Error:
	printf("get_bwg_descrpt: Failed memory alloc\n");
	return dcb_failed;

	CATCH_END(dcb_failed);
}

dcb_result put_bwg_descrpt(char *device_name, u8 bwgid, char *name)
{
	TRY_BEGIN;
	string dn(device_name); 
	dcb_result result = dcb_success;
	full_dcb_attribs attribs;
	unsigned int size;

	if (!name) return dcb_bad_params;
	size = (unsigned int)strlen(name);  /* Localization OK */

	memset(&attribs, 0, sizeof(attribs));
	bwg_desc_it it = bwg_desc.find(dn);
	
	if ((it != bwg_desc.end()) &&
		(bwgid < ((*it).second)->valid_desc_id_count)) {

		/* Only take as many characters as can be held */
		if (!(size < sizeof(((*it).second)->bwg_desc_of_id[bwgid])))
			size = sizeof(((*it).second)->bwg_desc_of_id[bwgid])-1;
		memcpy(((*it).second)->bwg_desc_of_id[bwgid], name, size);
		/* Put a null at the end incase it was truncated */
		((*it).second)->bwg_desc_of_id[bwgid][size] = '\0';
		if (set_persistent(device_name,NULL,NULL,
			((*it).second),NULL,0) != dcb_success) {
			return dcb_device_not_found;
		}
	} else {
		/* Store in persistent storage - though not in
		 * DCB data store */
		if (get_persistent(device_name, &attribs) == dcb_success) {
			if (!(size <
				sizeof(attribs.descript.bwg_desc_of_id[bwgid])))
				size = sizeof(
				 attribs.descript.bwg_desc_of_id[bwgid]) - 1;
			memcpy(attribs.descript.bwg_desc_of_id[bwgid],
				name, size);
			/* Put a null at the end incase it was
			 * truncated */
			attribs.descript.bwg_desc_of_id[bwgid][size] = '\0';
			if (set_persistent(device_name, NULL, NULL,
				&attribs.descript, NULL, 0) != dcb_success) {
				printf("Set persistent failed "
					   "in put_bwg_descrpt()\n");
				result = dcb_device_not_found;
			}
		}
		else
			result = dcb_device_not_found;
	}
	return result;

	CATCH_END(dcb_failed);
}

/******************************************************************************
**
** Method:      CopyConfigToOper
**
** Description: Function to copy local or peer PG or PFC or APPTLV
**              configurations to oper configuration.
**
** Arguments: char *device_name
**            u32 SrcFlag - Tells where to copy from (local or peer)
**            u32 EventFlag
**            u32 SubType - This is valid only for DCB_LOCAL_CHANGE_APPTLV and
**                          DCB_REMOTE_CHANGE_APPTLV
**
** Returns: TRUE if successful, failure code otherwise.
**
*******************************************************************************/
void CopyConfigToOper(char *device_name, u32 SrcFlag, u32 EventFlag,
			u32 Subtype)
{
	/* this function relies on the caller to acquire the DCB lock */
	TRY_BEGIN;
	string dn(device_name); 
	printf("CopyConfigToOper %s\n", device_name);
	if (DCB_TEST_FLAGS(EventFlag, DCB_LOCAL_CHANGE_PG, DCB_LOCAL_CHANGE_PG)
		||
		DCB_TEST_FLAGS(EventFlag, DCB_REMOTE_CHANGE_PG,
		DCB_REMOTE_CHANGE_PG)) {

		/* Get the Local or Peer store */
		pg_it Src;
		pg_it localSrc;
		if (SrcFlag == LOCAL_STORE) {
			Src = pg.find(dn);
			if (Src == pg.end())
				return;
		} else {
			if (SrcFlag == PEER_STORE) {
				Src = peer_pg.find(dn);
				if (Src == peer_pg.end())
					return;

				localSrc = pg.find(dn);
				if (localSrc == pg.end())
					return;
			} else {
				/* We don't support */
				return;
			}
		}
		/* Get the Oper store */
		pg_it Oper = oper_pg.find(dn);
		if (Oper == oper_pg.end()) {
			return;
		}
		/* Copy Src to Oper. */
		int i;

		for (i = 0; i < 8; i++) {
			(*Oper).second->tx.bwg_percent_of_link[i] =
			(*Src).second->tx.bwg_percent_of_link[i];
			(*Oper).second->rx.bwg_percent_of_link[i] =
			(*Src).second->rx.bwg_percent_of_link[i];

			(*Oper).second->tx.up[i].bandwidth_group_mapping =
			(*Src).second->tx.up[i].bandwidth_group_mapping;
			(*Oper).second->rx.up[i].bandwidth_group_mapping =
			(*Src).second->rx.up[i].bandwidth_group_mapping;

			(*Oper).second->tx.up[i].strict_priority =
			(*Src).second->tx.up[i].strict_priority;
			(*Oper).second->rx.up[i].strict_priority =
			(*Src).second->rx.up[i].strict_priority;

			(*Oper).second->tx.up[i].percent_of_bandwidth_group_cap =
			(*Src).second->tx.up[i].percent_of_bandwidth_group_cap;
			(*Oper).second->rx.up[i].percent_of_bandwidth_group_cap =
			(*Src).second->rx.up[i].percent_of_bandwidth_group_cap;

			if (SrcFlag == PEER_STORE) {
				(*Oper).second->tx.up[i].traffic_class_mapping =
					(*localSrc).second->tx.up[i].traffic_class_mapping;
				(*Oper).second->rx.up[i].traffic_class_mapping =
					(*localSrc).second->rx.up[i].traffic_class_mapping;
			} else {
				(*Oper).second->tx.up[i].traffic_class_mapping =
					(*Src).second->tx.up[i].traffic_class_mapping;
				(*Oper).second->rx.up[i].traffic_class_mapping =
					(*Src).second->rx.up[i].traffic_class_mapping;
			}
		}
	} else if (DCB_TEST_FLAGS(EventFlag, DCB_LOCAL_CHANGE_PFC,
		DCB_LOCAL_CHANGE_PFC) ||
		DCB_TEST_FLAGS(EventFlag, DCB_REMOTE_CHANGE_PFC,
		DCB_REMOTE_CHANGE_PFC)) {

		/* Get the Local or Peer store */
		pfc_it Src;
		if (SrcFlag == LOCAL_STORE) {
			Src = pfc.find(dn);
			if (Src == pfc.end())
				return;
		}
		else {
			if (SrcFlag == PEER_STORE) {
				Src = peer_pfc.find(dn);
				if (Src == peer_pfc.end())
					return;
			}
			else {
				/* We don't support */
				return;
			}
		}
		/* Get Oper store */
		pfc_it Oper = oper_pfc.find(dn);
		if (Oper == oper_pfc.end()) {
			return;
		}

		/* Copy Src to Oper. */
		memcpy( &(*Oper).second->admin, &(*Src).second->admin,
			sizeof((*Src).second->admin));

	} else if (DCB_TEST_FLAGS(EventFlag, DCB_LOCAL_CHANGE_APPTLV,
			DCB_LOCAL_CHANGE_APPTLV)||
			DCB_TEST_FLAGS(EventFlag, DCB_REMOTE_CHANGE_APPTLV,
			DCB_REMOTE_CHANGE_APPTLV)) {
		/* Get the Local or Peer store */
		/* Append subtype to the device name for APPTLV. */
		string dn1(device_name); 
		char sTemp[SHORT_STRING];
		snprintf(sTemp, SHORT_STRING, "%hu", Subtype);	  /* Localization OK */
		/* Append device name with subtype. */
		dn1.append(sTemp);

		app_it Src;
		if (SrcFlag == LOCAL_STORE) {
			Src = apptlv.find(dn1);
			if (Src == apptlv.end())
				return;
		} else {
			if (SrcFlag == PEER_STORE) {
				Src = peer_apptlv.find(dn1);
				if (Src == peer_apptlv.end())
					return;
			} else {
				/* We don't support */
				return;
			}
		}
		/* Get Oper store */
		app_it Oper = oper_apptlv.find(dn1);
		if (Oper != oper_apptlv.end()) {
			/* Copy Src to Oper. */
			printf("  Changing app data from %02x to %02x\n",
				(*Oper).second->AppData[0],
				(*Src).second->AppData[0]);
				(*Oper).second->Length = (*Src).second->Length;
			memcpy((*Oper).second->AppData, (*Src).second->AppData,
				(*Src).second->Length);
		}
	}
	CATCH_END_VOID;
}

/******************************************************************************
**
** Method:      LocalPeerCompatible
**
** Description: Function to check if local and peer configurations matches.
**
** Arguments: char *device_name
**            ULONG EventFlag
**            ULONG SubType - This is valid only for DCB_LOCAL_CHANGE_APPTLV
**			and DCB_REMOTE_CHANGE_APPTLV
**
** Returns: TRUE if successful, failure code otherwise.
**
*******************************************************************************/
boolean_t LocalPeerCompatible(char *device_name, u32 EventFlag, u32 Subtype)
{
	TRY_BEGIN;
	string dn(device_name); 
	if (DCB_TEST_FLAGS(EventFlag, DCB_LOCAL_CHANGE_PG, DCB_LOCAL_CHANGE_PG)
		||
		DCB_TEST_FLAGS(EventFlag, DCB_REMOTE_CHANGE_PG,
		DCB_REMOTE_CHANGE_PG)) {

		/* Get the Local and Peer PG store */
		pg_it Local = pg.find(dn);
		pg_attribs *lpg;
		pg_attribs *ppg;
		boolean_t match = FALSE;
		int i;

		if (Local == pg.end()) {
			goto Error;
		}
		pg_it Peer = peer_pg.find(dn);
		if (Peer == peer_pg.end()) {
			goto Error;
		}
		lpg = (*Local).second;
		ppg = (*Peer).second;

		match = TRUE;
		for (i = 0; i < 8; i++) {
			if (lpg->tx.bwg_percent_of_link[i] !=
				ppg->tx.bwg_percent_of_link[i])
				match = FALSE;
			if (lpg->tx.up[i].bandwidth_group_mapping !=
				ppg->tx.up[i].bandwidth_group_mapping)
				match = FALSE;
			if (lpg->tx.up[i].strict_priority !=
				ppg->tx.up[i].strict_priority)
				match = FALSE;
			if (lpg->tx.up[i].percent_of_bandwidth_group_cap !=
				ppg->tx.up[i].percent_of_bandwidth_group_cap)
				match = FALSE;
		}
		if (match) {
			printf("COMPAT PG - passed\n");
			return TRUE;
		}
		printf("COMPAT PG - failed\n");
	} else if (DCB_TEST_FLAGS(EventFlag, DCB_LOCAL_CHANGE_PFC,
			DCB_LOCAL_CHANGE_PFC)||
			DCB_TEST_FLAGS(EventFlag, DCB_REMOTE_CHANGE_PFC,
			DCB_REMOTE_CHANGE_PFC)) {

		/* Get the Local and Peer PFC store */
		pfc_it Local = pfc.find(dn);
		if (Local == pfc.end()) {
			goto Error;
		}
		pfc_it Peer = peer_pfc.find(dn);
		if (Peer == peer_pfc.end()) {
			goto Error;
		}
		if (memcmp(&(*Local).second->admin,
			&(*Peer).second->admin,
			sizeof((*Local).second->admin)) ==
			0 ) {
			printf("COMPAT PFC - passed\n");
			return TRUE;
		}

		printf("COMPAT PFC - failed\n");
	} else if (DCB_TEST_FLAGS(EventFlag, DCB_LOCAL_CHANGE_APPTLV,
			DCB_LOCAL_CHANGE_APPTLV)||
			DCB_TEST_FLAGS(EventFlag, DCB_REMOTE_CHANGE_APPTLV,
			DCB_REMOTE_CHANGE_APPTLV)) {

		/* Get the Local and Peer APPTLV store */
		/* Append subtype to the device name for APPTLV. */
		string dn1(device_name); 
		char sTemp[SHORT_STRING];
		snprintf(sTemp, SHORT_STRING, "%hu", Subtype);	  /* Localization OK */
		dn1.append(sTemp);
		app_it Local = apptlv.find(dn1);
		if (Local == apptlv.end()) {
			goto Error;
		}
		app_it Peer = peer_apptlv.find(dn1);
		if (Peer == peer_apptlv.end()) {
			goto Error;
		}
		if ((*Local).second->Length == (*Peer).second->Length) {
			if (memcmp((*Local).second->AppData,
				(*Peer).second->AppData,
				(*Local).second->Length) == 0) {
				return TRUE;
			}
		}
		printf("COMPAT APP - failed\n");
	} 
	return FALSE;
Error:
	printf("LocalPeerCompatible: device not found\n");
	return FALSE;

	CATCH_END(FALSE);
}

/* returns: 0 on success
 *          1 on failure
*/
int set_configuration(char *device_name, u32 EventFlag)
{
	TRY_BEGIN;
	string dn(device_name); 
	dcb_result sResult;

	if (DCB_TEST_FLAGS(EventFlag, DCB_LOCAL_CHANGE_PG, DCB_LOCAL_CHANGE_PG)
		||
		DCB_TEST_FLAGS(EventFlag, DCB_REMOTE_CHANGE_PG,
		DCB_REMOTE_CHANGE_PG)) {

		/* Get Oper store */
		pg_it Oper = oper_pg.find(dn);
		pg_it Local = pg.find(dn);
		if (Oper == oper_pg.end() || Local == pg.end()) {
			return 1;
		}
		pgroup_attribs pg_data;
		if (DCB_TEST_FLAGS(EventFlag, DCB_REMOTE_CHANGE_PG,
			DCB_REMOTE_CHANGE_PG)) {
			if ((sResult = dcb_check_config((*Oper).second))
				!= dcb_success) {
				printf("Rule check returned error %d\n",
					sResult);  /* Localization OK */
				return sResult;
			}
		}
		memcpy(&pg_data.rx, &(*Oper).second->rx, sizeof(pg_data.rx));
		memcpy(&pg_data.tx, &(*Oper).second->tx, sizeof(pg_data.tx));
		return set_hw_pg(device_name, &pg_data,
			(*Local).second->protocol.OperMode);

	} else if (DCB_TEST_FLAGS(EventFlag, DCB_LOCAL_CHANGE_PFC,
		DCB_LOCAL_CHANGE_PFC) ||
		DCB_TEST_FLAGS(EventFlag, DCB_REMOTE_CHANGE_PFC,
		DCB_REMOTE_CHANGE_PFC)) {

		/* Get Oper store */
		pfc_it Oper = oper_pfc.find(dn);
		pfc_it Local = pfc.find(dn);
		if (Oper == oper_pfc.end() || Local == pfc.end()) {
			return 1;
		}
		return set_hw_pfc(device_name, (*Oper).second->admin,
			(*Local).second->protocol.OperMode);
	}
	return 0;

	CATCH_END(1);
}

/******************************************************************************
**
** Method:      run_feature_protocol
**
** Description: This function runs feature state machine for a local or remote
** change.
** The function caller should acquire lock before calling this function.
** Caller must call this function per event.
**
** Arguments: char *device_name
**            u32 EventFlag
**            u32 SubType - This is valid only for DCB_LOCAL_CHANGE_APPTLV and
**				DCB_REMOTE_CHANGE_APPTLV
**
** Returns: dcb_success if successful, failure code otherwise.
**
*******************************************************************************/
dcb_result run_feature_protocol(char *device_name, u32 EventFlag, u32 Subtype)
{
	feature_protocol_attribs *feature_protocol      = NULL;
	feature_protocol_attribs *peer_feature_protocol = NULL;
	boolean_t ErrorChanged  = FALSE;
	boolean_t Err, local_change;
	pg_attribs oper_pg_before;
	int save_pg_opermode = 0;
	u32 pg_events = 0;
	pfc_attribs oper_pfc_before;
	int save_pfc_opermode = 0;
	u32 pfc_events = 0;
	app_attribs oper_app_before;
	int save_app_opermode = 0;
	u32 app_events = 0;
	

	memset(&oper_pg_before,0,sizeof(pg_attribs));
	memset(&oper_pfc_before,0,sizeof(pfc_attribs));
	memset(&oper_app_before,0,sizeof(app_attribs));

	TRY_BEGIN;
	string dn(device_name); 
	/* Get the protocol store */
	if (DCB_TEST_FLAGS(EventFlag, DCB_EVENT_FLAGS,
		DCB_LOCAL_CHANGE_PG) ||
		DCB_TEST_FLAGS(EventFlag, DCB_EVENT_FLAGS,
		DCB_REMOTE_CHANGE_PG)) {

		/* Get the local feature protocol */
		pg_it it = pg.find(dn);
		if (it != pg.end()) {
			feature_protocol = (&((*it).second)->protocol);
			save_pg_opermode = feature_protocol->OperMode;
		} else {
			goto ErrNoDevice;
		}

		/* Get the remote feature protocol */
		pg_it it1 = peer_pg.find(dn);
		if (it1 != peer_pg.end()) {
			peer_feature_protocol =
				(&((*it1).second)->protocol);
		} else {
			goto ErrNoDevice;
		}

		pg_it Oper = oper_pg.find(dn);
		if (Oper != oper_pg.end()) {
			memcpy(&oper_pg_before.rx, &(*Oper).second->rx,
				sizeof(oper_pg_before.rx));
			memcpy(&oper_pg_before.tx, &(*Oper).second->tx,
				sizeof(oper_pg_before.tx));
		} else {
			goto ErrNoDevice;
		}
	}

	if (DCB_TEST_FLAGS(EventFlag, DCB_EVENT_FLAGS,
		DCB_LOCAL_CHANGE_PFC) ||
		DCB_TEST_FLAGS(EventFlag, DCB_EVENT_FLAGS,
		DCB_REMOTE_CHANGE_PFC)) {

		/* Get the local feature protocol */
		pfc_it it = pfc.find(dn);
		if (it != pfc.end()) {
			feature_protocol = (&((*it).second)->protocol);
			save_pfc_opermode = feature_protocol->OperMode;
		} else {
			goto ErrNoDevice;
		}

		/* Get the remote feature protocol */
		pfc_it it1 = peer_pfc.find(dn);
		if (it1 != peer_pfc.end()) {
			peer_feature_protocol =
				(&((*it1).second)->protocol);
		} else {
			goto ErrNoDevice;
		}

		pfc_it Oper = oper_pfc.find(dn);
		if (Oper != oper_pfc.end()) {
			memcpy(&oper_pfc_before.admin,
				&(*Oper).second->admin,
				sizeof(oper_pfc_before.admin));
		} else {
			goto ErrNoDevice;
		}
	}
	if (DCB_TEST_FLAGS(EventFlag, DCB_EVENT_FLAGS,
		DCB_LOCAL_CHANGE_APPTLV) ||
		DCB_TEST_FLAGS(EventFlag, DCB_EVENT_FLAGS,
		DCB_REMOTE_CHANGE_APPTLV)) {

		string dn1(device_name); 
		char sTemp[SHORT_STRING];
		snprintf(sTemp, SHORT_STRING, "%hu", Subtype);	  /* Localization OK */
		dn1.append(sTemp);
		/* Get the local feature protocol */
		app_it it = apptlv.find(dn1);
		if (it != apptlv.end()) {
			feature_protocol = (&((*it).second)->protocol);
			save_app_opermode = feature_protocol->OperMode;

		} else {
			goto ErrNoDevice;
		}

		/* Get the remote feature protocol */
		app_it it1 = peer_apptlv.find(dn1);
		if (it1 != peer_apptlv.end()) {
			peer_feature_protocol =
				(&((*it1).second)->protocol);
		} else {
			goto ErrNoDevice;
		}

		app_it Oper = oper_apptlv.find(dn1);
		if (Oper != oper_apptlv.end()) {
			oper_app_before.Length = (*Oper).second->Length;
			memcpy(&oper_app_before.AppData[0],
				&((*Oper).second->AppData[0]),
				oper_app_before.Length);
		} else {
			goto ErrNoDevice;
		}
	}

	/* Get the local control protocol variables. */
	control_prot_it control_protocol = dcb_control_prot.find(dn);
	if (control_protocol == dcb_control_prot.end()) {
		goto ErrNoDevice;
	}
	/* Get the remote control protocol variables. */
	control_prot_it peer_control_protocol =
		dcb_peer_control_prot.find(dn);
	if (peer_control_protocol == dcb_peer_control_prot.end()) {
		goto ErrNoDevice;
	}
	if ((feature_protocol == NULL) ||
		(peer_feature_protocol == NULL)) {
		goto ErrNoDevice;
	}

	if (feature_protocol->State == DCB_INIT) {
		feature_protocol->Oper_version =
			feature_protocol->Max_version;
		feature_protocol->OperMode = FALSE;
		feature_protocol->Error = FALSE;

		/* Set the parameters. */
		feature_protocol->FeatureSeqNo =
			((*control_protocol).second)->SeqNo + 1;
		/* If Syncd false, then control state machine will
		 * TX LLDP message with local config. */
		feature_protocol->Syncd =
			!(feature_protocol->Advertise);
		printf("  Set Syncd to %u [%u]\n",
			feature_protocol->Syncd, __LINE__);
		feature_protocol->State = DCB_LISTEN;
	}
	if (feature_protocol->State == DCB_LISTEN) {
		printf("Feature state machine\n");
		local_change = FALSE;
		if (DCB_TEST_FLAGS(EventFlag, DCB_EVENT_FLAGS,
			DCB_LOCAL_CHANGE_PG) ||
			DCB_TEST_FLAGS(EventFlag, DCB_EVENT_FLAGS,
			DCB_LOCAL_CHANGE_PFC) ||
			DCB_TEST_FLAGS(EventFlag, DCB_EVENT_FLAGS,
			DCB_LOCAL_CHANGE_APPTLV)) {

			local_change = TRUE;
			printf("  Local change:");
			printf(" %s",
				(EventFlag & DCB_LOCAL_CHANGE_PG) ?
				"PG" : "");
			printf(" %s",
				(EventFlag & DCB_LOCAL_CHANGE_PFC) ?
				"PFC" : "");
			printf(" %s",
				(EventFlag & DCB_LOCAL_CHANGE_APPTLV) ?
				"APP" : "");
			printf("\n");
		}

		/* If local changed and we are already synched... */
		if (local_change && feature_protocol->Syncd) {
			printf("  Local feature already synced\n");
			/* If we are not synched, we won't be able
			 * to accept new local changes until we get
			 * back remote changes for previous local
			 * change. */

			/* Set the parameters. */
			feature_protocol->FeatureSeqNo =
				((*control_protocol).second)->SeqNo + 1;
			/* If Syncd false, then control state machine
			 * will TX LLDP message with local config. */
			if ((feature_protocol->Advertise == TRUE) ||
				(feature_protocol->Advertise_prev ==
				TRUE)) {
				feature_protocol->Syncd = FALSE;
				printf("  Set Syncd to %u [%u]\n",
					feature_protocol->Syncd, __LINE__);
			} else {
				feature_protocol->Syncd = TRUE;
				printf("  Set Syncd to %u [%u]\n",
					feature_protocol->Syncd, __LINE__);
				feature_protocol->tlv_sent = TRUE;
			}
		}
		/* F4 If don't advertise, then copy the local config to
		 * Oper config. */
		if (!feature_protocol->Advertise) {
			printf("F5 - Advertise mode OFF:");
			printf(" %s", (EventFlag&(DCB_LOCAL_CHANGE_PG |
				DCB_REMOTE_CHANGE_PG)) ? "PG" : "");
			printf(" %s", (EventFlag&(DCB_LOCAL_CHANGE_PFC |
				DCB_REMOTE_CHANGE_PFC)) ? "PFC" : "");
			printf(" %s",
				(EventFlag&(DCB_LOCAL_CHANGE_APPTLV |
				DCB_REMOTE_CHANGE_APPTLV)) ? "APP" :"");
			printf("\n");

			/* copy the local config to Oper config. */
			CopyConfigToOper(device_name, LOCAL_STORE,
				EventFlag, Subtype);
			/* State already in Listen so don't have to
			 * change. */
			feature_protocol->Error = FALSE;
			feature_protocol->Error_Flag = 0;
			feature_protocol->OperMode =
				feature_protocol->Enable;
			set_configuration(device_name, EventFlag);
			goto OperChange;
		}

		/* Process remote change. */
		if (DCB_TEST_FLAGS(EventFlag, DCB_EVENT_FLAGS,
			DCB_REMOTE_CHANGE_PG) ||
			DCB_TEST_FLAGS(EventFlag, DCB_EVENT_FLAGS,
			DCB_REMOTE_CHANGE_PFC)||
			DCB_TEST_FLAGS(EventFlag, DCB_EVENT_FLAGS,
			DCB_REMOTE_CHANGE_APPTLV)) {

			printf("  Remote change: ");
			printf(" %s",
				(EventFlag & DCB_REMOTE_CHANGE_PG) ?
				"PG" : "");
			printf(" %s",
				(EventFlag & DCB_REMOTE_CHANGE_PFC) ?
				"PFC" : "");
			printf(" %s", (EventFlag &
				DCB_REMOTE_CHANGE_APPTLV) ? "APP" : "");
			printf("\n");

			/* This version check is part of the Control
			 * protocol state machine that  must be
			 * complete before proceeding with feature
			 * protocol state machine */
			if (!((control_protocol->second->Oper_version ==
				min(peer_control_protocol->second->Max_version,
				control_protocol->second->Max_version)) &&
				(control_protocol->second->Oper_version ==
				peer_control_protocol->second->Oper_version)
				)) {
				goto ErrBadVersion;
			}

			/* F7 If feature TLV not present then Opermode false. */
			if (!peer_feature_protocol->RxFeatureTLVPresent) {
				printf("F8 - Feature not present\n");
				/* copy the local config to Oper config. */
				CopyConfigToOper(device_name, LOCAL_STORE,
					EventFlag, Subtype);
				feature_protocol->OperMode = FALSE;
				feature_protocol->Syncd = TRUE;
				printf("  Set Syncd to %u [%u]\n",
					feature_protocol->Syncd, __LINE__);
				feature_protocol->Oper_version =
					feature_protocol->Max_version;
				feature_protocol->Error = FALSE;
				feature_protocol->Error_Flag = 0x20;
				set_configuration(device_name, EventFlag);
				goto OperChange;
			} else {
				feature_protocol->Error_Flag &= ~0x20;
			}
			if (!feature_protocol->Syncd &&
				(((*peer_control_protocol).second)->AckNo
				!= feature_protocol->FeatureSeqNo)) {

				/* Wait for the Peer to synch up. */
				printf("  Wait for Peer to synch: "
					"Peer AckNo %d, FeatureSeqNo %d \n",
					((*peer_control_protocol).second)->AckNo,
					feature_protocol->FeatureSeqNo);
				goto OperChange;
			}
			/* Check for the Oper version */
			if (feature_protocol->Oper_version !=
				min(peer_feature_protocol->Max_version,
				feature_protocol->Max_version)) {

				/* Update Oper version and signal
				 *LLDP send. */
				feature_protocol->Oper_version =
					min(peer_feature_protocol->Max_version,
					feature_protocol->Max_version);
				printf("  Update feature oper version to %d and"
					" signal send\n",
					feature_protocol->Oper_version);
				feature_protocol->Syncd = FALSE;
				printf("  Set Syncd to %u [%u]\n",
					feature_protocol->Syncd, __LINE__);
				feature_protocol->FeatureSeqNo =
					((*control_protocol).second)->SeqNo + 1;
				goto OperChange;
			}
			feature_protocol->Syncd =  TRUE;
			printf("  Set Syncd to %u [%u]\n",
				feature_protocol->Syncd, __LINE__);
				/* F13/F14 */

			if (feature_protocol->Oper_version !=
				peer_feature_protocol->Oper_version ) {
				/* Wait for Peer to synch up with
				 * version. */
				printf("  Wait for the Peer to synch up ");
				printf("with feature version.\n");
				goto OperChange;
			}

			feature_protocol->PeerWilling =
				peer_feature_protocol->Willing;
			/* F15 If feature is disabled on any side,
			 * then make Opermode false. */
			if (!feature_protocol->Enable ||
				!peer_feature_protocol->Enable) {
				printf("F16 - Feature is disabled\n");
				/* Copy Local config to Oper config. */
				CopyConfigToOper(device_name,
					LOCAL_STORE, EventFlag, Subtype);
				feature_protocol->OperMode = FALSE;
				feature_protocol->Error_Flag = 0;
				Err = feature_protocol->Error;
				/* Set_configuration to driver. */
				feature_protocol->Error =
					(set_configuration(device_name,
					EventFlag) != dcb_success);
				if (Err != feature_protocol->Error) {
					ErrorChanged = TRUE;
				}
				if (feature_protocol->Error)
					feature_protocol->Error_Flag |= 0x2;
				goto ErrProt;
			}
			/* F17 */
			if (feature_protocol->Willing &&
				!feature_protocol->PeerWilling) {
				printf("F18 - local willing,  "
					"peer NOT willing\n");
				/* Copy Peer config to Oper config. */
				CopyConfigToOper(device_name,
					PEER_STORE, EventFlag, Subtype);
				feature_protocol->OperMode = TRUE;
				feature_protocol->Error_Flag = 0;
				Err = feature_protocol->Error;
				/* Set_configuration to driver. */
				feature_protocol->Error =
					(set_configuration(device_name,
					EventFlag) != dcb_success);
				if (Err != feature_protocol->Error) {
					ErrorChanged = TRUE;
				}
				if (feature_protocol->Error)
					feature_protocol->Error_Flag |= 0x2;
				goto ErrProt;
			}
			/* F19 */
			if (!feature_protocol->Willing &&
				feature_protocol->PeerWilling) {
				printf("F20 - local NOT willing,  "
					"peer willing\n");
				/* Copy Local config to Oper config. */
				CopyConfigToOper(device_name,
					LOCAL_STORE, EventFlag, Subtype);
				feature_protocol->OperMode = TRUE;
				feature_protocol->Error_Flag = 0;
				Err = feature_protocol->Error;
				/* Set_configuration to driver. */
				feature_protocol->Error =
					(set_configuration(device_name,
					EventFlag) != dcb_success);
				if (Err != feature_protocol->Error)
					ErrorChanged = TRUE;
				if (feature_protocol->Error)
					feature_protocol->Error_Flag |= 0x2;
				goto ErrProt;

			}
			/* F21 */
			if ((feature_protocol->Willing ==
				feature_protocol->PeerWilling) &&
				(LocalPeerCompatible(device_name,
				EventFlag, Subtype))) {
				printf("F22 - local willing,  peer willing\n");
				/* Copy Local config to Oper config. */
				CopyConfigToOper(device_name,
					LOCAL_STORE, EventFlag, Subtype);
				feature_protocol->OperMode = TRUE;
				feature_protocol->Error_Flag = 0;
				Err = feature_protocol->Error;
				/* Set_configuration to driver. */
				feature_protocol->Error =
					(set_configuration(device_name,
					EventFlag) != dcb_success);
				if (Err != feature_protocol->Error)
					ErrorChanged = TRUE;
				if (feature_protocol->Error)
					feature_protocol->Error_Flag |= 0x2;
			} else {
				printf("F23 - Local & Peer config not"
					" compatible\n");
				printf("  LocalPeerCompatible error\n");
				/* Copy Local config to Oper config. */
				CopyConfigToOper(device_name,
					LOCAL_STORE, EventFlag, Subtype);
				feature_protocol->OperMode = FALSE;
				if (!feature_protocol->Error)
					ErrorChanged =  TRUE;
				/* Set default configuration */
				if (set_configuration(device_name, EventFlag)
					!= dcb_success)
					feature_protocol->Error_Flag |= 0x2;
				feature_protocol->Error = TRUE;
				feature_protocol->Error_Flag |= 0x1;
			}
ErrProt:
			if (peer_feature_protocol->Error)
				feature_protocol->Error_Flag |= 0x8;
			if (feature_protocol->Error ||
				peer_feature_protocol->Error ) {
				printf("  ## FEATURE ERROR: "
					"%d, %d (Error_Flag 0x%x EventFlag 0x%x)\n",
					feature_protocol->Error,
					peer_feature_protocol->Error,
					feature_protocol->Error_Flag,
					EventFlag);
				if (feature_protocol->OperMode) {
					feature_protocol->OperMode = FALSE;
					/* Set default configuration */
					set_configuration(device_name,
							EventFlag);
				}
			}
			if (ErrorChanged) {
				printf("  ErrorChanged \n");
				feature_protocol->Syncd = FALSE;
				printf("  Set Syncd to %u [%u]\n",
					feature_protocol->Syncd, __LINE__);
				feature_protocol->FeatureSeqNo =
					((*control_protocol).second)->SeqNo+ 1;
			}
		}
OperChange:
		if (DCB_TEST_FLAGS(EventFlag, DCB_EVENT_FLAGS,
			DCB_LOCAL_CHANGE_PG) ||
			DCB_TEST_FLAGS(EventFlag, DCB_EVENT_FLAGS,
			DCB_REMOTE_CHANGE_PG)) {

			pg_it Oper = oper_pg.find(dn);
			if (Oper == oper_pg.end()) {
				goto ErrNoDevice;
			}
			if (memcmp(&(oper_pg_before.tx),
				&((*Oper).second->tx),
				sizeof(oper_pg_before.tx)) != 0)
				pg_events = pg_events | EVENT_OPERATTR;
			if (feature_protocol->OperMode != save_pg_opermode)
				pg_events = pg_events | EVENT_OPERMODE;
		}
		if (DCB_TEST_FLAGS(EventFlag, DCB_EVENT_FLAGS,
			DCB_LOCAL_CHANGE_PFC)||
			DCB_TEST_FLAGS(EventFlag, DCB_EVENT_FLAGS,
			DCB_REMOTE_CHANGE_PFC)) {

			pfc_it Oper = oper_pfc.find(dn);
			if (Oper == oper_pfc.end()) {
				goto ErrNoDevice;
			}
			if (memcmp(&(oper_pfc_before.admin),
				&((*Oper).second->admin),
				sizeof(oper_pfc_before.admin)) != 0)
				pfc_events = pfc_events | EVENT_OPERATTR;

			if (feature_protocol->OperMode !=
				save_pfc_opermode)
				pfc_events = pfc_events | EVENT_OPERMODE;
		}
		if (DCB_TEST_FLAGS(EventFlag, DCB_EVENT_FLAGS,
			DCB_LOCAL_CHANGE_APPTLV) ||
			DCB_TEST_FLAGS(EventFlag, DCB_EVENT_FLAGS,
			DCB_REMOTE_CHANGE_APPTLV)) {

			string dn1(device_name);
			char sTemp[SHORT_STRING];
			/* Localization OK */
			snprintf(sTemp, SHORT_STRING, "%hu", Subtype);
			dn1.append(sTemp);

			app_it Oper = oper_apptlv.find(dn1);
			if (Oper == oper_apptlv.end()) {
				goto ErrNoDevice;
			}
			if ((oper_app_before.Length != (*Oper).second->Length)
				|| (oper_app_before.Length &&
				(memcmp(oper_app_before.AppData,
				(*Oper).second->AppData,
				oper_app_before.Length) != 0))) {
				app_events = app_events|EVENT_OPERATTR;
			}

			if (feature_protocol->OperMode != save_app_opermode)
				app_events = app_events | EVENT_OPERMODE;
		}
	}
	CATCH_END(dcb_failed);

	if (pg_events) {
		pg_event(device_name, pg_events);
	}
	if (pfc_events) {
		pfc_event(device_name, pfc_events);
	}
	if (app_events) {
		app_event(device_name, Subtype, app_events);
	}
	return dcb_success;

ErrNoDevice:
	return dcb_device_not_found;

ErrBadVersion:
	printf("  Versions not compatible\n");
	return dcb_ctrl_vers_not_compatible;
}


dcb_result GetDCBTLVState(char *device_name, u8 *State)
{
	TRY_BEGIN;
	string dn(device_name); 
	/* Get the remote control protocol variables. */
	control_prot_it peer_control_protocol = dcb_peer_control_prot.find(dn);
	if (peer_control_protocol == dcb_peer_control_prot.end()) {
		return dcb_device_not_found;
	}
	*State = (u8)peer_control_protocol->second->RxDCBTLVState;

	return dcb_success;

	CATCH_END(dcb_failed);
}

void SendLLDPMessage(char *device_name, boolean_t NewFeatureTLVs )
{
	/* Create DCB TLV and append stored feature TLVs to it. */
	/* send LLDP message. */
	/* lldp_send(char *device_name, void *tlv); */
	process_somethingChangedLocal(device_name, NewFeatureTLVs);
}

boolean_t FeaturesSynched(char *device_name)
{
	TRY_BEGIN;
	string dn(device_name); 

	/* Get the local PG feature protocol */
	pg_it it = pg.find(dn);
	if (it == pg.end()) {
		return FALSE;
	}
	if ((&((*it).second)->protocol)->State == DCB_LISTEN) {
		if ((&((*it).second)->protocol)->Syncd == FALSE ||
			(&((*it).second)->protocol)->tlv_sent == FALSE)
			return FALSE;
	}

	/* Get the local PFC feature protocol */
	pfc_it it1 = pfc.find(dn);
	if (it1 == pfc.end()) {
		return FALSE;
	}
	if ((&((*it1).second)->protocol)->State == DCB_LISTEN) {
		if ((&((*it1).second)->protocol)->Syncd == FALSE ||
			(&((*it1).second)->protocol)->tlv_sent == FALSE)
			return FALSE;
	}

	/* Get the APP TLV feature protocol. */
	for (int i = 0; i < DCB_MAX_APPTLV ; i++) {
		string dn1(device_name); 
		char sTemp[SHORT_STRING];
		 /* Localization OK */
		snprintf(sTemp, SHORT_STRING, "%hu", i);
		/* Append device name with subtype. */
		dn1.append(sTemp);
		app_it it2 = apptlv.find(dn1);
		if (it2 == apptlv.end()) {
			return FALSE;
		}
		if ((&((*it2).second)->protocol)->State == DCB_LISTEN) {
			if ((&((*it2).second)->protocol)->Syncd == FALSE ||
				(&((*it2).second)->protocol)->tlv_sent == FALSE)
				return FALSE;
		}
	}
	return TRUE;

	CATCH_END(FALSE);
}

/* Set the Syncd value to TRUE for features which are not advertising.
*/
void update_feature_syncd(char *device_name)
{
	TRY_BEGIN;
	string dn(device_name); 

	/* Get the local PG feature protocol */
	pg_it it = pg.find(dn);
	if (it != pg.end())
		if ((&((*it).second)->protocol)->Advertise == FALSE)
			(&((*it).second)->protocol)->Syncd = TRUE;

	/* Get the local PFC feature protocol */
	pfc_it it1 = pfc.find(dn);
	if (it1 != pfc.end())
		if ((&((*it1).second)->protocol)->Advertise == FALSE)
			(&((*it1).second)->protocol)->Syncd = TRUE;

	/* Get the APP TLV feature protocol. */
	for (int i = 0; i < DCB_MAX_APPTLV ; i++) {
		string dn1(device_name);
		char sTemp[SHORT_STRING];
		/* Localization OK */
		snprintf(sTemp, SHORT_STRING, "%hu", i);
		/* Append device name with subtype. */
		dn1.append(sTemp);
		app_it it2 = apptlv.find(dn1);
		if (it2 == apptlv.end()) {
			return;
		}
		if ((&((*it2).second)->protocol)->Advertise == FALSE) {
			(&((*it2).second)->protocol)->Syncd = TRUE;
			printf("  Set Syncd to %u [%u]\n",
				(&((*it2).second)->protocol)->Syncd, __LINE__);
		}
	}

	CATCH_END_VOID;

}
/******************************************************************************
**
** Method:      run_control_protocol
**
** Description: This function runs control state machine for a local or
**		remote change.
** The function caller should acquire lock before calling this function.
** Caller must call this function for local or remote change but not both.
**
** Arguments: char *device_name
**            u32 EventFlag
** Returns: dcb_success if successful, failure code otherwise.
**
*******************************************************************************/
dcb_result run_control_protocol(char *device_name, u32 EventFlag)
{
	TRY_BEGIN;
	string dn(device_name);

	/* Get the local control protocol variables. */
	control_prot_it control_protocol = dcb_control_prot.find(dn);
	if (control_protocol == dcb_control_prot.end()) {
		return dcb_device_not_found;
	}
	/* Get the remote control protocol variables. */
	control_prot_it peer_control_protocol = dcb_peer_control_prot.find(dn);
	if (peer_control_protocol == dcb_peer_control_prot.end()) {
		return dcb_device_not_found;
	}

	if (control_protocol->second->State == DCB_INIT) {
		/* Set the parameters. */
		control_protocol->second->SeqNo =  0;
		control_protocol->second->AckNo = 0;
		control_protocol->second->MyAckNo = 0;
		control_protocol->second->Oper_version = 
					control_protocol->second->Max_version;
		control_protocol->second->State = DCB_LISTEN;
	}
	if (control_protocol->second->State == DCB_LISTEN) {
		printf("DCB Ctrl in LISTEN \n");
		/* Process local change if any. */
		if (DCB_TEST_FLAGS(EventFlag, DCB_LOCAL_CHANGE_PG,
			DCB_LOCAL_CHANGE_PG) ||
			DCB_TEST_FLAGS(EventFlag, DCB_LOCAL_CHANGE_PFC,
			DCB_LOCAL_CHANGE_PFC)||
			DCB_TEST_FLAGS(EventFlag, DCB_LOCAL_CHANGE_APPTLV,
			DCB_LOCAL_CHANGE_APPTLV)) {

			printf("  Local change detected: ");
			printf(" %s", (EventFlag & DCB_LOCAL_CHANGE_PG) ?
				"PG" : "");
			printf(" %s", (EventFlag & DCB_LOCAL_CHANGE_PFC) ?
				"PFC" : "");
			printf(" %s", (EventFlag & DCB_LOCAL_CHANGE_APPTLV) ?
				"APP" : "");
			printf("\n");

			if (control_protocol->second->SeqNo ==
				control_protocol->second->MyAckNo) {
				printf("  Local SeqNo == Local AckNo\n");
				if (!FeaturesSynched(device_name)) {
					update_feature_syncd(device_name);
					control_protocol->second->SeqNo++;

					printf("  *** Sending packet -- ");
					printf("SeqNo = %d \t AckNo =  %d \n",
						control_protocol->second->SeqNo,
						control_protocol->second->AckNo);

					/* Send new DCB ctrl & feature TLVs */
					SendLLDPMessage(device_name, TRUE);
				}
			}
			return dcb_success;
		}
		/* Process remote change if any. */
		if (DCB_TEST_FLAGS(EventFlag, DCB_REMOTE_CHANGE_PG,
			DCB_REMOTE_CHANGE_PG) ||
			DCB_TEST_FLAGS(EventFlag, DCB_REMOTE_CHANGE_PFC,
			DCB_REMOTE_CHANGE_PFC) ||
			DCB_TEST_FLAGS(EventFlag, DCB_REMOTE_CHANGE_APPTLV,
			DCB_REMOTE_CHANGE_APPTLV)) {

			boolean_t SendDCBTLV = FALSE;
			printf("  Remote change detected: ");
			printf(" %s", (EventFlag & DCB_REMOTE_CHANGE_PG) ?
				"PG" : "");
			printf(" %s", (EventFlag & DCB_REMOTE_CHANGE_PFC) ?
				"PFC" : "");
			printf(" %s", (EventFlag & DCB_REMOTE_CHANGE_APPTLV) ?
				"APP" : "");
			printf("\n");

			u8 State;
			if (GetDCBTLVState(device_name, &State) ==
				dcb_success) {
				if (State == DCB_PEER_EXPIRED) {
					control_protocol->second->SeqNo = 0;
					control_protocol->second->AckNo = 0;
					control_protocol->second->MyAckNo = 0;
					control_protocol->second->Oper_version =
					  control_protocol->second->Max_version;
					peer_control_protocol->second->RxDCBTLVState =
						DCB_PEER_RESET;

					/* Send the updated DCB TLV */
					SendDCBTLV = TRUE;
					printf("  Peer expired\n");
					goto send;
				}
			}

			if (control_protocol->second->Oper_version !=
				min(peer_control_protocol->second->Max_version,
				control_protocol->second->Max_version)) {

				control_protocol->second->Oper_version =
					min(peer_control_protocol->second->Max_version,
					control_protocol->second->Max_version);

				/* Send the updated DCB TLV */
				SendDCBTLV = TRUE;
				printf("  Change Oper Version \n");
				goto send;
			}

			if (control_protocol->second->Oper_version !=
				peer_control_protocol->second->Oper_version) {
				/* Wait for peer to synch up. */
				printf("  Wait for Peer to synch \n");
				goto send;
			}
			/* Update MyAck */
			control_protocol->second->MyAckNo =
				peer_control_protocol->second->AckNo;

			/* If received new Peer TLV, then acknowledge the
			 * Peer TLV */
			if (control_protocol->second->AckNo !=
				peer_control_protocol->second->SeqNo) {
				control_protocol->second->AckNo =
					peer_control_protocol->second->SeqNo;
				SendDCBTLV = TRUE;
			}

			/* If changes in feature then send message with latest
			 * DCB and FeatureTLV */
send:
			printf("  Current -- SeqNo = %d \t MyAckNo =  %d \n",
				control_protocol->second->SeqNo,
				control_protocol->second->MyAckNo);
			if ((control_protocol->second->SeqNo ==
				control_protocol->second->MyAckNo)  &&
				(!FeaturesSynched(device_name))) {
				printf("  Features not synced \n");
				update_feature_syncd(device_name);
				/* Send LLDP message. */
				control_protocol->second->SeqNo++;
				printf("  *** Sending Packet -- ");
				printf("SeqNo = %d \t AckNo =  %d \n",
					control_protocol->second->SeqNo,
					control_protocol->second->AckNo);
				/* Send new DCB control & feature TLVs*/
				SendLLDPMessage(device_name, TRUE);
				return dcb_success;
			}

			if (SendDCBTLV) {
				printf("  SendDCBTLV is set \n");
				/* if you didn't send LLDP message above then
				 * send one without changing feature TLVs. */
				printf("  *** Sending Packet -- ");
				printf("SeqNo = %d \t AckNo =  %d \n",
					control_protocol->second->SeqNo,
					control_protocol->second->AckNo);
				/* Send new DCB TLVs with old feature TLVs. */
				SendLLDPMessage(device_name, FALSE);
			}
		}
	}
	return dcb_success;

	CATCH_END(dcb_failed);
}

/******************************************************************************
**
** Method:      run_dcb_protocol
**
** Description: This function runs both feature and control state machines
**	for the features that are specified in the event flag. The function
**	caller should acquire lock per port before calling this function.
**	Caller can only club together local  PG*PFC*APPTLV or
**	remote PG*PFC*APPTLV eventflags and call this function.
**
** Arguments: char *device_name
**            u32 EventFlag
**            u32 SubType - This is valid for APPTLV event flags only.
**                          If >= DCB_MAX_APPTLV, then we process all Subtypes
**				for APPTLV flags.
** Returns: dcb_success if successful, failure code otherwise.
**
*******************************************************************************/
dcb_result run_dcb_protocol(char *device_name, u32 EventFlag, u32 Subtype)
{
	dcb_result result = dcb_success;
	boolean_t LocalChange = FALSE;
	u32 i, SubTypeMin, SubTypeMax;

	printf("running DCB protocol for %s, flags:%04x\n", device_name,
		EventFlag);

	/* if valid use SubType param, otherwise process all SubTypes */
	if (Subtype < DCB_MAX_APPTLV) {
		SubTypeMin = Subtype;
		SubTypeMax = Subtype+1;
	}
	else {
		SubTypeMin = 0;
		SubTypeMax = DCB_MAX_APPTLV;
	}
	/* Run the feature state machines. */
	if (DCB_TEST_FLAGS(EventFlag, DCB_LOCAL_CHANGE_PG, DCB_LOCAL_CHANGE_PG)
		&& (result != dcb_ctrl_vers_not_compatible)) {
		result = run_feature_protocol(device_name, DCB_LOCAL_CHANGE_PG);
		LocalChange = TRUE;
	}
	if (DCB_TEST_FLAGS(EventFlag, DCB_LOCAL_CHANGE_PFC,
		DCB_LOCAL_CHANGE_PFC)
		&& (result != dcb_ctrl_vers_not_compatible)) {
		result = run_feature_protocol(device_name,
			DCB_LOCAL_CHANGE_PFC);
		LocalChange = TRUE;
	}
	if (DCB_TEST_FLAGS(EventFlag, DCB_LOCAL_CHANGE_APPTLV,
		DCB_LOCAL_CHANGE_APPTLV) &&
		(result != dcb_ctrl_vers_not_compatible)) {

		for (i = SubTypeMin; i < SubTypeMax; i++) {
			result = run_feature_protocol(device_name,
				DCB_LOCAL_CHANGE_APPTLV, i);
		}
		LocalChange = TRUE;
	}
	/* Only allow local or remote change at a time. */
	if (!LocalChange) {
		if (DCB_TEST_FLAGS(EventFlag, DCB_REMOTE_CHANGE_PG,
			DCB_REMOTE_CHANGE_PG)
			&& (result != dcb_ctrl_vers_not_compatible)) {
			result = run_feature_protocol(device_name,
				DCB_REMOTE_CHANGE_PG);
		}
		if (DCB_TEST_FLAGS(EventFlag, DCB_REMOTE_CHANGE_PFC,
			DCB_REMOTE_CHANGE_PFC)
			&& (result != dcb_ctrl_vers_not_compatible)) {
			result = run_feature_protocol(device_name,
				DCB_REMOTE_CHANGE_PFC);
		}
		if (DCB_TEST_FLAGS(EventFlag, DCB_REMOTE_CHANGE_APPTLV,
			DCB_REMOTE_CHANGE_APPTLV)
			&& (result != dcb_ctrl_vers_not_compatible)) {
			for (i = SubTypeMin; i < SubTypeMax; i++) {
				result = run_feature_protocol(device_name,
					DCB_REMOTE_CHANGE_APPTLV, i);
			}
		}
	}

	/* apply all feature setting to the driver: linux only */
	set_hw_all(device_name);

	/* Run the control state machine. */
	result = run_control_protocol(device_name, EventFlag);

	return result;
}
